#*************************************************************************************************************
# Name: Sandbox
# Author: Elxx (http://elxx.net)
# Version: BF2142 Edition 0.2
#*************************************************************************************************************
# Conditions and Limitations: You may use this file locally, on listening and dedicated servers.
# This software is provided "as is". In no event shall the author, authors or contributors be held
# liable for  any direct, indirect, incidental, special, exemplary, or consequential damages (including)
# but not limited to, procurement of substitute good or services; loss of use, data or profits; or
# business interruption) however caused and on any theory of liability, whether in contract, strict liability
# arising in any way out of the use of this software, even if advised of the possibility of such damage.
#*************************************************************************************************************
#
# Configuration of Sandbox admins has been moved to sandbox.con in the mod's Settings directory.
#
#*************************************************************************************************************

import host
import bf2
import bf2.Timer
import bf2.PlayerManager
import math
import sys
import re
import os
import inspect
#import sandboxUploader
import time
#from bf2.stats.constants import *

class pData:
	def __init__(self):
		self.tracerDistance = 3
		self.objects = []
		self.objectTypes = []
		self.grabbed = []
		self.lastRotation = 0
		self.savedRotation = ()
		self.lastSpawned = []
		self.lastTemplate = ""
		self.tracerOn = 1
		self.tracerDistance = 8
		self.gridSnap = 0
		self.rotSnap = 5
		self.autoGrab = 1
		self.numSpawned = [0,0]
		self.preGrab = []
		self.preGrabMode = 0
		self.lastAction = []
		self.spawnDelay = 0
		self.grabMode = 3

# Grab Modes
# 1 = Original, own objects
# 2 = Original, all objects
# 3 = Extra-Accuracy, own objects
# 4 = Extra-Accuracy, all objects

# Undo Actions
# 1 = grab [object, original position, original rotation]
# 2 = drop [object, frozen]
# 3 = delete [object, original position] (disabled)
# 4 =

def ExceptionOutput():
	sys.stderr.write("\n" + "Exception Occured: " + str(sys.exc_info()[0]) + "\n")
	sys.stderr.write("Value: " + str(sys.exc_info()[1]) + "\n")
	sys.stderr.write("Line:" + str(readline(inspect.getfile(sys.exc_info()[2]),
							sys.exc_info()[2].tb_lineno)) + "\n")
	sys.stderr.write("Line #: " + str(sys.exc_info()[2].tb_lineno) + "\n")
	sys.stderr.write("File: " + str(inspect.getfile(sys.exc_info()[2])) + "\n" + "\n")
	host.rcon_invoke("game.sayAll \"Exception Caught: " + str(sys.exc_info()[1]) + "\"")
	host.rcon_invoke("game.sayAll \"Line " + str(sys.exc_info()[2].tb_lineno) + "\"")

def readline(filename, lineno):
	filen = re.sub('\\\\', '/', filename)
	file = open(filen, 'rU')
	lines = file.readlines()
	file.close()
	linen = lineno - 1
	line = re.sub('\s+', ' ', lines[linen])
	return line

SANDBOX_VERSION = "0.4"

class sandboxCore:
	def __init__(self):
		self.SANDBOX_VERSION = SANDBOX_VERSION
		# Set up event handlers
		host.registerGameStatusHandler(self.onGameStatusChanged)
		host.registerHandler('PlayerDeath', self.onPlayerDeathCQ)
		host.registerHandler('PlayerKilled', self.onPlayerKilledCQ)
		host.registerHandler('PlayerSpawn', self.onPlayerSpawn)
		host.registerHandler('EnterVehicle', self.onEnterVehicle)
		host.registerHandler('ExitVehicle', self.onExitVehicle)
		host.registerHandler('PlayerDisconnect', self.onPlayerDisconnect)
		host.registerHandler('ChatMessage', self.onChatMessage, 1)
		host.registerHandler('RemoteCommand', self.onRemoteCommand, 1)

	def initSbx(self):
		# All allowed static objects (not including vehicles)
		# I split this up to easily keep track of the sources for each list.
		#self.allowed_objects = ['container_empty_c','container_seald_d','fence_industry_16m','fence_corrugated_8m','crate01_c','barrel_big_c']
		self.allowed_objects = ['beacon_tower_01_ay', 'refugee_divider_a', 'refugee_grate_solid_a', 'refugee_grate_stairs_a', 'refugee_marquee_a', 'refugeemodule_01_a', 'refugeemodule_01_a_90degcol', 'refugeemodule_01_b', 'refugeemodule_01_b_90degcol', 'refugeemodule_01_c', 'refugeemodule_01_c_90degcol', 'refugeemodule_01x8tall_a', 'tollstation_02_a', '', '_hothouse_part01_a', '_hothouse_part02_a', '_hothouse_part03_a', 'bc_big_bridge_ay', 'bc_building_a', 'bc_yard_a', 'freezer', 'freezer_cluster1', 'freezer_cluster2', 'irregation_entrance_a', 'irregation_plant_main_a', 'irregation_plant_top', 'irrigation_pipeline_30degrees_a', 'irrigation_pipeline_60m_a', 'irrigation_pipeline_90curve_10m', 'irrigation_pipeline_90curve_20m', 'irrigation_pipeline_120m_a', 'irrigation_pipeline_support_a', 'irrigation_plantpod_a', 'reactor_foundation_a', 'reactor_generator_a', 'reactor_powerreserve_a', 'ambstat_ACSmoke', 'ambstat_Battlesmoke', 'ambstat_civwreck_burning', 'ambstat_Clouds', 'ambStat_Damn_waterFall', 'ambstat_Electric', 'ambstat_Fire', 'ambstat_Fire_medium', 'ambstat_Fire_small', 'ambstat_Light_Red2', 'ambstat_Light_Red', 'ambstat_MassiveSmoke', 'ambstat_OnlyFire', 'ambstat_OnlySmoke', 'amdStat_Damn_WaterImpact', 'amdStat_Fountain_waterSpray', 'ground_3x6', 'ground_6x6', 'ground_8x4', 'ground_10x2_5', 'ground_16x4', 'ground_16x8', 'wall_2x4', 'wall_3x6', 'wall_6x3', 'wall_6x12', 'wall_8x4', 'wall_8x8', 'wall_8x16', 'wall_12x24', 'barrack_double_c', 'barrack_single_c', 'barrack_single_c_component', 'com_tower_c', 'guard_watchtower_01_c', 'guardwall_6m_broken_c', 'guardwall_12m_broken_c', 'guardwall_12m_c', 'guardwall_12m_door_c', 'guardwall_12m_gate_c', 'guardwall_12m_railinghole_c', 'guardwall_24m_c', 'guardwall_45deg_c', 'guardwall_90deg_c', 'guardwall_doordoor_c', 'guardwall_gatedoor_c', 'guardwall_tjunction_c', 'office_01_c', 'park_monument_01', 'park_monument_02', 'warehouse_large_01_c', 'warehouse_medium_01_c', 'warehouse_small_01_c', 'connector', 'barrel_big_dest_c', 'crate01_dest_c', 'eu_mech_wreck_01_c', 'alarm_device', 'alarm_device_compleat', 'antenna_01_c', 'barge', 'barrel_big_a', 'barrel_big_b', 'barrel_big_c', 'barrel_small_a', 'barrel_small_b', 'barrel_small_c', 'basketball_basket', 'brickwall_4m_c', 'brickwall_4m_wreck_c', 'brickwall_8m_c', 'brickwall_16m_c', 'brickwall_32m_c', 'brickwall_64m_c', 'brickwall_door_8m_c', 'brickwall_end_c', 'bridge_industrial_01_c', 'cable12m', 'cable16m', 'cable20m', 'cable24m', 'cable30m', 'cable40m', 'cable40mtwo', 'cable60m', 'cable60mtwo', 'catwalk_2_cross_c', 'catwalk_3_cross_c', 'catwalk_4_cross_c', 'catwalk_8m_c', 'catwalk_16m_c', 'catwalk_24m_c', 'catwalk_32m_c', 'catwalk_40m_c', 'catwalk_90deg_c', 'catwalk_cluster1_c', 'catwalk_cluster1_mirror_c', 'catwalk_cluster2_c', 'catwalk_end_poles', 'catwalk_stair_c', 'catwalk_support_c', 'catwalk_tower_c', 'catwalk_tower_wall_c', 'cistern_building_c', 'cistern_medium_c', 'cistern_small_c', 'concrete_wall_10m_stair_in_c', 'concrete_wall_10m_stair_out_c', 'concrete_wall_16m_c', 'concrete_wall_32m_c', 'concrete_wall_64m_c', 'concrete_wall_corner_10m_c', 'concrete_wall_corner_10m_inv_c', 'concrete_wall_inner_slope_30m_c', 'concrete_wall_inner_slope_30m_inv_c', 'concrete_wall_slope_30m_c', 'concrete_wall_slope_30m_inv_c', 'concwall_4m_c', 'concwall_4m_wreck_c', 'concwall_8m_c', 'concwall_16m_c', 'concwall_32m_c', 'concwall_64m_c', 'concwall_door_8m_c', 'concwall_end_c', 'container_cluster2_c', 'container_empty_a', 'container_empty_b', 'container_empty_c', 'container_empty_d', 'container_empty_e', 'container_seald_a', 'container_seald_b', 'container_seald_c', 'container_seald_d', 'container_seald_e', 'container_small_c', 'container_wreck_c', 'crane_small_c', 'crate01_a', 'crate01_b', 'crate01_c', 'crate01_stack01_a', 'crate01_stack01_b', 'crate01_stack01_c', 'crate01_stack02_a', 'crate01_stack02_b', 'crate01_stack02_c', 'crate01_stack03_a', 'crate01_stack03_b', 'crate01_stack03_c', 'crate01_stack04_a', 'crate01_stack04_b', 'crate01_stack04_c', 'fence_corrugated_8m', 'fence_corrugated_16m', 'fence_corrugated_32m', 'fence_corrugated_64m', 'fence_corrugated_broken_4m', 'fence_corrugated_broken_8m', 'fence_corrugated_broken_16m', 'fence_industry_8m', 'fence_industry_16m', 'fence_industry_32m', 'fence_industry_64m', 'fence_industry_broken_4m', 'fence_industry_broken_8m', 'fence_industry_broken_16m', 'floodlight_big_c', 'gangplank_01', 'gangplank_02', 'gastank_c', 'harbor_edge_10m_c', 'harbor_edge_20m_45deg_c', 'harbor_edge_20m_c', 'harbor_edge_40m_c', 'harbor_edge_45degree_in2_c', 'harbor_edge_45degree_in_c', 'harbor_edge_45degree_out2_c', 'harbor_edge_45degree_out_c', 'harbor_edge_80m_c', 'harbor_edge_90degree_in_c', 'harbor_edge_90degree_out_c', 'harbor_edge_ramp_c', 'industry_vehicle_01_c', 'ladder_4m_c', 'ladder_4m_irgsupport', 'ladder_5m_c', 'ladder_5m_irgsupport', 'ladder_6m_c', 'ladder_7m_beams_c', 'ladder_9m_beams_c', 'ladder_10m_beams_c', 'ladder_10m_c', 'ladder_10m_irgsupport', 'ladder_12m_cistmedium_c', 'ladder_20m_beams_c', 'ladder_20m_c', 'ladder_20m_irgsupport', 'ladder_40m_c', 'ladder_40m_irgsupport', 'lamp_post_double_c', 'lamp_post_single_c', 'largeflowerpot_01_c', 'metalbeam_stack_01_c', 'monorail_01', 'monorail_01_slope', 'monorail_02', 'monorail_02_v2', 'monorail_03', 'monorail_platform_c', 'monorail_sign_01_c', 'monorail_sign_02_c', 'monorail_train_c', 'monorail_train_wreck_c', 'parkbench_01_c', 'parkbench_02_c', 'parked_car_01_c', 'parked_car_02_c', 'parked_truck_01_c', 'pedestrian_railing_8m_broken_c', 'pedestrian_railing_8m_c', 'pedestrian_railing_16m_c', 'pedestrian_railing_32m_c', 'playground_horse', 'playground_swing_c', 'residetial_divider2_c', 'residetial_divider_c', 'road_sign_01_c', 'road_sign_02_c', 'road_sign_03_c', 'rockpile_large', 'rockpile_medium', 'rockpile_small', 'ship_01_c', 'solarpanel_01_c', 'solarpanel_02_c', 'stone_stairs_5x10m_c', 'stonewall_low_4mend_c', 'stonewall_low_4mwreck_c', 'stonewall_low_8m_c', 'stonewall_low_16m_c', 'stormdrain', 'train_industial_c', 'train_industial_wagon_c', 'train_track_16m_end_c', 'train_track_32m_c', 'train_track_64m_c', 'train_track_y_32m_c', 'trash_dumpster_01_c', 'ventilation_duct', 'as_apc_wreck', 'as_at_wreck', 'as_heavy_mech_wreck', 'as_tank_wreck', 'eu_ag_wreck', 'barn_01_v01_e', 'farmhouseold_01_v01_e', 'farmhouseold_02_v01_e', 'farmhouseold_03_v01_e', 'farmhouseold_04_v01_e', 'farmshed_01_v01_e', 'farmshed_02_v01_e', 'residential_house_01_e', 'residential_house_02_e', 'residential_house_03_e', 'residential_house_03_ruin_e', 'residential_house_04_e', 'ruinedhouse_01_e', 'ruinedhouse_03_e', 'scraper_01_e', 'scraper_base_e', 'scraper_stairs_e', 'thouseold_01_ver01_e', 'thouseold_01_ver02_e', 'thouseold_02_ruin01_e', 'thouseold_02_v01_e', 'thouseold_03_ruin01_e', 'thouseold_03_v01_e', 'thouseold_03_v02_e', 'citylight_01', 'citylight_02', 'statue_01_e', 'statue_01_e_v2', 'statue_02_e', 'wastebasket_01', 'church', 'command_center_e', 'flowerscape_sidewalk', 'flowerscrape_large', 'flowerscrape_medium', 'flowerscrape_ruin_01', 'flowerscrape_ruin_02', 'flowerscrape_small', 'flowerscrape_tower_l', 'flowerscrape_tower_r', 'controltower_01_m', 'mil_base_house_2', 'pilbox_smal_01_m', 'pilbox_smal_01_m', 'pilbox_smal_01_m', 'pilbox_smal_01_m', 'crate_ammo_01_m', 'helipad_m', 'infanterycover_3m_m', 'infanterycover_14m_m', 'infanterycover_broken_7m_m', 'infanterycover_u_m', 'tank_obstacle_four_m', 'tank_obstacle_single_m', 'tank_obstacle_three_m', 'trench_20m', 'trench_90degree', 'trench_bridge', 'trench_bridge_small', 'trench_end', 'trench_exit', 'trench_t_crossing', 'flag_EU', 'flag_neutral', 'flag_Pac', 'flagpole', 'pac_light_soldier', 'pac_heavy_soldier', 'us_light_soldier', 'us_heavy_soldier']
		# Let's add weapons and stuff too!
		self.allowed_objects += ['As_aa', 'as_ar_rifle', 'as_ar_rocket', 'as_ar_shotgun', 'As_av', 'as_handgun', 'as_mg', 'as_smg', 'as_sni', 'Beacon', 'EU_AA', 'eu_ar_rifle', 'EU_AR_Rocket', 'EU_AR_Shotgun', 'EU_AV', 'EU_Handgun', 'eu_mg', 'EU_SMG', 'eu_sni', 'Hud_As_Assault', 'Hud_As_Engineer', 'Hud_As_Recon', 'Hud_As_Support', 'Hud_Eu_Assault', 'Hud_Eu_Engineer', 'Hud_Eu_Recon', 'Hud_Eu_Support', 'Item_Dummy', 'Item_Lock', 'item_Lock_Big', 'Knife', 'Knife_unlock', 'ordermarker', 'ParachuteLauncher', 'sentrygun_deployer', 'Unl_Active_Camo', 'unl_adv_sni', 'unl_ammo_upgrade', 'Unl_AV_Rifle', 'unl_best_buy_rifle', 'unl_best_buy_rocket', 'unl_best_buy_shotgun', 'Unl_c4', 'unl_c4_detonator', 'Unl_Carbine', 'Unl_Defibrillator', 'Unl_Defuser', 'Unl_Deployer_Sentrygun', 'unl_drone_recon', 'Unl_Drone_Recon_Detonator', 'unl_drone_sentry', 'Unl_Drone_Sentry_Detonator', 'Unl_Grenade_EMP', 'Unl_Grenade_Frag', 'Unl_Grenade_Smoke', 'Unl_har_rifle', 'unl_har_rocket', 'unl_har_shotgun', 'Unl_HmgSupportUnlock-Ganz', 'Unl_Hub_Ammo', 'Unl_Hub_Medic', 'Unl_lar_rifle', 'Unl_lar_rocket', 'unl_lar_shotgun', 'unl_medic_upgrade', 'Unl_Mine_AntiPersonnel', 'Unl_Mine_EMP', 'Unl_Mine_Motion', 'unl_port_shield', 'Unl_Pulse_Meter', 'unl_repair', 'unl_repair_upgrade', 'unl_scope_zoom', 'Unl_Shotgun', 'unl_sonar', 'unl_stabilizer', 'as_drone', 'commander_EMP_strike', 'commander_movement_analyst', 'commander_orbital_Strike', 'commander_supplybox', 'commander_uav_drone', 'commander_uav_drone_control', 'eu_drone', 'eu_sentrygun', 'portable_shield', 'spawn_beacon', 'unl_hub_ammo_stationary']
		# Fire!
		self.allowed_objects += ['ambstat_fire_medium','glow2']
		# Static objects defined as "annoying" will have a double delay before subsequent spawns
		self.annoying_objects = []
		# Static objects that are multiplayer-compatible, that is, "_mp" can be appended to the name to
		# get a networked version of the object
		self.mp_objects = []
		# Objects that simply don't work in multiplayer
		self.nonmp_objects = []
		# Spawnable vehicles, will not be auto-grabbed
		self.allowed_vehicles = ['as_ag', 'as_apc', 'as_at', 'as_fav', 'as_heavy_mech', 'as_tank', 'as_titan', 'eu_apc', 'eu_at', 'eu_fav', 'eu_pod', 'EU_TANK', 'eu_titan', 'static_aa', 'static_av', 'us_ag', 'us_heavy_mech']
		# Out of the previous, the ones that will be auto-grabbed if the option is enabled
		self.autograb_excempt = ['usaas_stinger','usaas_stinger_spw','ats_tow']
		# Soldiers? Erm, k...
		self.soldier_templates = []
		# -------------------------
		self.object_types = []
		self.deleted_objects = []
		self.obj_freeze = {}
		self.obj_locked = {}
		self.spawned_objects = []
		self.is_spGame = 1
		self.sbxTimer = None
		self.freezerTimer = None
		self.p_objects = {}
		self.allow_lock = True
		self.allow_grab = True
		self.allow_create = True
		self.gravity = -15
		self.spawnerTemplates = []
		self.lastObject = []
		self.allowCreateOverride = True
		self.sBoundaries = True
		self.pdata = {}
		self.preCompletedObjects = []
		self.relative_grabbing = True
		self.admins = ""
		self.lastObjectCounter = 0
		# Set up the sandbox timer
		self.timerCounter = 0
		if not self.sbxTimer:
			self.sbxTimer = bf2.Timer(self.run_sbxTimer, 0 , 1)
			self.sbxTimer.setRecurring(0.01)
		if not self.freezerTimer:
			self.freezerTimer = bf2.Timer(self.freezer, 0 , 1)
			self.freezerTimer.setRecurring(0.01)
		# Debug mode? This prints a bunch of stuff to stdout so you can log it with a console or something
		self.enableDebug = 1
		self.dbg("Sandbox has been initialized.")

	def onGameStatusChanged(self, status):
		if status == bf2.GameStatus.Playing:
			self.preCompletedObjects = []
			for x in self.allowed_objects:
				for y in bf2.objectManager.getObjectsOfTemplate(x):
					self.preCompletedObjects.append(x)
			for x in bf2.objectManager.getObjectsOfType('dice.hfe.world.ObjectTemplate.PlayerControlObject'):
				self.preCompletedObjects.append(x)
			g_controlPoints = bf2.objectManager.getObjectsOfType('dice.hfe.world.ObjectTemplate.ControlPoint')
			self.con_save_cp = 1
			try:
				for obj in g_controlPoints:
					if obj.cp_getParam('team') == 2:
						self.con_save_cp = obj.getTemplateProperty('ControlPointId')
					elif obj.cp_getParam('team') == 0:
						obj.cp_setParam('team', 2)
			except:
				pass
			self.p_objects = {}
			self.pdata = {}
			self.object_types = []
			self.deleted_objects = []
			self.obj_freeze = {}
			self.obj_locked = {}
			self.gravity = -15
			self.spawnerTemplates = []
			self.spawned_objects = []
			self.ftpUpload = False
			self.ftpOptions = {
				'server': '',
				'username': '',
				'password': '',
				'target': '',
				'dlLocation': '',
			}
			host.rcon_invoke("physics.gravity " + str(self.gravity))
			if self.sBoundaries:
				self.enableBoundaries()
			else:
				self.disableBoundaries()
			try:
				admin_file = open(host.sgl_getModDirectory() + "/settings/sandbox.con")
				print admin_file.read().split("\n")
				admin_file.seek(0)
				nil,self.admins,nil,self.ftpUpload,nil,self.ftpOptions['server'],nil,self.ftpOptions['username'],nil,self.ftpOptions['password'],nil,self.ftpOptions['target'],nil,self.ftpOptions['dlLocation'] = admin_file.read().split("\n")
				self.ftpUpload = bool(self.ftpUpload)
			except:
				ExceptionOutput()
				self.admins = ""
			print self.admins
			self.dbg(self.admins)
			self.dbg("Game status cleared and Sandbox reloaded.")

	def onRemoteCommand(self, playerId, cmd):
		from math import pi

		if playerId == -1: #fix for local servers
			playerId = 255
		self.dbg("Command received: [" + str(cmd) + "] from [" + str(playerId) + "]")
		p = bf2.PlayerManager.Player(playerId)
		pD = self.pdata[p.getName()]
		pPos = p.getDefaultVehicle().getPosition()
		pRot = p.getDefaultVehicle().getRotation()
		origCmd = cmd
		spacepos = origCmd.find(' ')
		if spacepos == -1: spacepos=len(cmd)
		cmd = origCmd[0:spacepos]
		cmdVar = origCmd[spacepos + 1:]
		if cmd == "btnSpawn":
			if not self.allow_create and not (self.checkAdmin(p.getName()) and self.allowCreateOverride):
				return 0
			if cmdVar and (not self.allowed_objects.count(cmdVar.lower()) and not self.allowed_vehicles.count(cmdVar.lower())):
				return 0
			if pD.spawnDelay > 0:
				return 0
			if self.nonmp_objects.count(cmdVar.lower()) and not self.is_spGame:
				return 0
			if self.mp_objects.count(cmdVar.lower()) and not self.is_spGame:
				cmdVar += "_mp"
			pCam = self.getSoldierCam(p)
			pCamRot = pCam.getRotation()
			sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
			sRot = pRot
			self.dbg("new position: " + str(sPos[0]) + "," + str(sPos[1]) + "," + str(sPos[2]))
			self.dbg("new rotation: " + str(sRot[0]) + "," + str(sRot[1]) + "," + str(sRot[2]))
			if cmdVar:
				pD.lastTemplate = str(cmdVar)
			else:
				if pD.lastRotation:
					sRot = pD.lastRotation
			cmdVar = pD.lastTemplate
			try:
				#self.createSpawnerObject(str(cmdVar), sPos, sRot, p.index)
				self.createObject(str(cmdVar), sPos, sRot, p.index)
			except:
				ExceptionOutput()
			if pD.objectTypes.count(str(cmdVar)) < 1:
				pD.objectTypes.append(str(cmdVar))
			if self.object_types.count(str(cmdVar)) < 1:
				self.object_types.append(str(cmdVar))
			pD.lastRotation = sRot
			if not self.is_spGame:
				if self.annoying_objects.count(str(cmdVar)):
					pD.spawnDelay = 120
				else:
					pD.spawnDelay = 80
			self.dbg("Created spawner!")
			print time.strftime("%H:%M:%S") + " - [" + p.getName() + "] - [" + str(cmdVar) + "]"

		if cmd == "btnGrabAlt":
			if self.is_spGame:
				if pD.preGrabMode == 0:
					pD.preGrabMode = 1
				else:
					o = pD.preGrab[0]
					if hasattr(o,"oId"):
						host.rcon_invoke("Object.active id" + str(o.oId))
						host.rcon_invoke("Object.isVisible 1")
					pD.preGrab[0].setPosition(pD.preGrab[3])
					pD.preGrabMode = 0
					pD.preGrab = []

		if cmd == "btnGrab":
			if pD.spawnDelay > 40:
				return 0
			if not self.allow_grab and not self.checkAdmin(p.getName()):
				return 0
			if pD.preGrabMode == 1 and self.is_spGame:
				o = pD.preGrab[0]
				if hasattr(o,"oId"):
					host.rcon_invoke("Object.active id" + str(o.oId))
					host.rcon_invoke("Object.isVisible 1")
				pD.preGrab[0].setPosition(pD.preGrab[3])
				pD.preGrabMode = 0
				pD.preGrab = []
			shortest = [0,pD.tracerDistance + 20.0]
			pCam = self.getSoldierCam(p)
			pCamRot = pCam.getRotation()
			sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
			sRot = pRot
			#for v in self.object_types:
			if pD.grabMode == 2 or pD.grabMode == 4:
				for v in pD.objectTypes:
					for q in bf2.objectManager.getObjectsOfTemplate(v):
						if q.isValid():
							qDist = self.getVectorDistance(sPos,q.getPosition())
							if qDist < shortest[1]:
								shortest = [q,qDist]
			else:
				for q in pD.objects:
					if q.isValid():
						qDist = self.getVectorDistance(sPos,q.getPosition())
						if qDist < shortest[1]:
							shortest = [q,qDist]
			if shortest[0]:
				if not self.is_spGame and self.obj_locked.has_key(shortest[0]) and not self.checkAdmin(p.getName()):
					if self.obj_locked[shortest[0]] != p.getName(): return 0
				self.dbg("Found closest object")
				distance = self.getVectorDistance(pPos,shortest[0].getPosition())
				frozen = 0
				if self.obj_freeze.has_key(shortest[0]):
					frozen = 1
					self.obj_freeze.pop(shortest[0])
				if pD.grabMode <= 2:
					pD.grabbed = [shortest[0],distance,1,shortest[0].getPosition(),shortest[0].getRotation()]
				else:
					shp = shortest[0].getPosition()
					sPos = self.getVectoredPosition(pPos,distance,pRot,pCamRot)
					grabOffset = ((sPos[0] - shp[0]),(sPos[1] - shp[1]),(sPos[2] - shp[2]))
					pD.grabbed = [shortest[0],distance,1,shortest[0].getPosition(),shortest[0].getRotation(),grabOffset]
				pD.lastAction.append([1,shortest[0],shortest[0].getPosition(),shortest[0].getRotation(),frozen])
				# 1 = grab [object, original position, original rotation, original frozen]
				self.dbg("Grabbed closest object")
				if not self.is_spGame:
					pD.spawnDelay = 80

		if cmd == "btnGrabMode":
			if pD.grabMode == 1:
				pD.grabMode = 2
				self.sendMsg(p, "Original [All Objects] grab mode selected")
			elif pD.grabMode == 2:
				pD.grabMode = 3
				pD.grabbed = []
				self.sendMsg(p, "Extra-Accuracy [My Objects] grab mode selected")
			elif pD.grabMode == 3:
				pD.grabMode = 4
				self.sendMsg(p, "Extra-Accuracy [All Objects] grab mode selected")
			elif pD.grabMode == 4:
				pD.grabMode = 1
				pD.grabbed = []
				self.sendMsg(p, "Original [My Objects] grab mode selected")

		if cmd == "btnRotate":
			v = cmdVar.split(" ")
			if pD.grabbed:
				if v[1] == "1": rFactor = pD.rotSnap
				else: rFactor = -pD.rotSnap
				obj = pD.grabbed[0]
				if obj and obj.isValid():
					objRot = obj.getRotation()
					if v[0] == "A":
						newRot = (objRot[0] + rFactor, objRot[1], objRot[2])
					if v[0] == "P":
						newRot = (objRot[0], objRot[1] + rFactor, objRot[2])
					if v[0] == "R":
						newRot = (objRot[0], objRot[1], objRot[2] + rFactor)
					pD.grabbed[4] = newRot
					obj.setRotation(newRot)
					pD.lastRotation = newRot

		if cmd == "btnResetRotation":
			if pD.grabbed:
				obj = pD.grabbed[0]
				if obj and obj.isValid():
					if not pD.savedRotation:
						pD.grabbed[4] = (0.0,0.0,0.0)
						obj.setRotation((0.0,0.0,0.0))
					else:
						pD.grabbed[4] = pD.savedRotation
						obj.setRotation(pD.savedRotation)

		if cmd == "btnSaveRotation":
			if not pD.savedRotation:
				if pD.grabbed:
					obj = pD.grabbed[0]
					if obj and obj.isValid():
						pD.savedRotation = obj.getRotation()
			else:
				pD.savedRotation = ()
				self.sendMsg(p, "Saved rotation reset cleared.")

		if cmd == "btnDrop":
			pD.lastAction.append([2,pD.grabbed[0],pD.grabbed[1],0])
			# 2 = drop [object, distance]
			pD.grabbed = []

		if cmd == "btnFreeze":
			if pD.grabbed:
				if pD.grabbed[0].isValid():
					self.obj_freeze[pD.grabbed[0]] = [pD.grabbed[0].getPosition(),pD.grabbed[0].getRotation()]
					pD.lastAction.append([2,pD.grabbed[0],pD.grabbed[1],1])
					pD.grabbed = []

		if cmd == "btnLock":
			if pD.grabbed and self.allow_lock:
				if pD.grabbed[0].isValid():
					self.obj_locked[pD.grabbed[0]] = p.getName()

		if cmd == "btnUnLock":
			if pD.grabbed:
				if pD.grabbed[0].isValid():
					try:
						del self.obj_locked[pD.grabbed[0]]
					except:
						pass #do nothing

		if cmd == "btnDelete":
			if pD.grabbed:
				if pD.grabbed[0].isValid():
					o = pD.grabbed[0]
					pD.grabbed = []
					self.deleteObject(o)

		if cmd == "btnPush":
			if cmdVar == "1":
				pD.grabbed[1] += 0.5
			else:
				pD.grabbed[1] += 1.0

		if cmd == "btnPull":
			if cmdVar == "1":
				pD.grabbed[1] -= 0.5
			else:
				pD.grabbed[1] -= 1.0

		if cmd == "btnUndo":
			#if pD.lastAction and self.is_spGame:
			if pD.lastAction:
				a = pD.lastAction.pop()
				if a[0] == 1: #grab [object, original position, original rotation, original frozen]
					try:
						self.dbg("Grab undoing")
						o = a[1]
						if pD.grabbed[0] == o:
							pD.grabbed = []
						o.setPosition(a[2])
						o.setRotation(a[3])
						if a[4] or (not self.is_spGame):
							self.obj_freeze[o] = [a[2],a[3]]
							self.dbg("Refrozen")
					except:
						ExceptionOutput()
				if a[0] == 2: # 2 = drop [object, distance]
					try:
						self.dbg("Drop undoing")
						o = a[1]
						pD.grabbed = []
						pD.grabbed = [o,a[2],1,o.getPosition(),o.getRotation(),(0.0,0.0,0.0)]
					except:
						ExceptionOutput()
				if a[0] == 3: # 3 = delete [object name, original position, original rotation]
					self.dbg("Delete undoing")

					#o = a[1]
					#o.setPosition(a[2])
					#self.deleted_objects.remove(o)
					#pD.numSpawned[0] += 1
					#self.onRemoteCommand(playerId,"btnUndo") #Undo once more so we re-grab the deleted object
				self.dbg("Undone.")

		if cmd == "btnTracerToggle":
			children = p.getPrimaryWeapon().getChildren()
			for child in children:
				if child.templateName == "glowshit":
					pos = child.getPosition()
					if pD.tracerOn:
						child.setPosition((pos[0],pos[1] - 100.0,pos[2]))
						pD.tracerOn = 0
					else:
						child.setPosition((pos[0],pos[1] + 100.0,pos[2]))
						pD.tracerOn = 1

		if cmd == "btnAutoGrabToggle":
			if pD.autoGrab:
				pD.autoGrab = 0
				self.sendMsg(p,"Autograb disabled.")
			else:
				pD.autoGrab = 1
				self.sendMsg(p,"Autograb enabled.")

		if cmd == "btnTracerDist":
			if cmdVar == "push":
				pD.tracerDistance += 1
			else:
				pD.tracerDistance -= 1
			self.sendMsg(p, "Tracer distance set to " + str(pD.tracerDistance))
			if self.is_spGame:
				children = p.getPrimaryWeapon().getChildren()
				for child in children:
					if child.templateName == "glowshit":
						pPos = p.getPrimaryWeapon().getPosition()
						pCam = self.getSoldierCam(p)
						pCamRot = pCam.getRotation()
						sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
						child.setPosition(sPos)

		if cmd == "btnRotSnap":
			if cmdVar == "inc":
				if pD.rotSnap > 4:
					pD.rotSnap += 5
				else:
					pD.rotSnap += 1
			else:
				if pD.rotSnap > 5:
					pD.rotSnap -= 5
				elif pD.rotSnap <= 5 and pD.rotSnap > 1:
					pD.rotSnap -= 1
			self.sendMsg(p, "Rotation factor set to " + str(pD.rotSnap) + "deg")

		if cmd == "btnSnap":
			if cmdVar == "inc":
				if pD.gridSnap < 0.25: pD.gridSnap += 0.05
				else: pD.gridSnap += 0.25
			else:
				if pD.gridSnap <= 0.25:
					if pD.gridSnap > 0:
						pD.gridSnap -= 0.05
				else:
					pD.gridSnap -= 0.25
			if pD.gridSnap < 0.05: pD.gridSnap = 0.0
			if pD.gridSnap > 0:
				self.sendMsg(p, "Grid snap set to " + str(pD.gridSnap))
			else:
				self.sendMsg(p, "Grid snap disabled")

		if self.checkAdmin(p.getName()) or self.is_spGame == 1:

			if cmd == "btnGravity" and cmdVar:
				try:
					gravFactor = int(cmdVar)
					self.gravity += gravFactor
					self.sendMsg(0,"Gravity set to " + str(self.gravity))
					host.rcon_invoke("physics.gravity " + str(self.gravity))
				except:
					ExceptionOutput()

			if cmd == "btnAdGrab" and cmdVar:
				if cmdVar == "1":
					self.allow_grab = True
					self.sendMsg(0,"Object grab enabled")
				else:
					self.allow_grab = False
					self.sendMsg(0,"Object grab disabled")

			if cmd == "btnAdLock" and cmdVar:
				if cmdVar == "1":
					self.allow_lock = True
					self.sendMsg(0,"Object lock enabled")
				else:
					self.allow_lock = False
					self.sendMsg(0,"Object lock disabled")

			if cmd == "btnAdCreate" and cmdVar:
				if cmdVar == "1":
					self.allow_create = True
					self.sendMsg(0,"Object create enabled")
				else:
					self.allow_create = False
					self.sendMsg(0,"Object create disabled")

			if cmd == "delAll" and cmdVar:
				try:
					if self.pdata.has_key(cmdVar):
						plD = self.pdata[cmdVar]
						plD.grabbed = []
						for o in plD.objects:
							self.deleteObject(o)
						self.sendMsg(0, str(cmdVar) + " - All objects deleted")
				except:
					ExceptionOutput()
			elif cmd == "delAll":
				pD.grabbed = []
				for o in pD.objects:
					self.deleteObject(o)

		if cmd == "btnNone":
			self.dbg("btnNone")

	#######################################
	#         Sandbox core timers         #
	#######################################
	#######################################

	def run_sbxTimer(self, data):
		try:
			self.timerCounter += 1
			if self.timerCounter >= 5:
				self.timerCounter = 0
			for p in bf2.playerManager.getPlayers():
				if not self.p_objects.has_key(p.getName()):
					self.p_objects[p.getName()] = p
				if not self.pdata.has_key(p.getName()):
					self.pdata[p.getName()] = pData()
			if self.lastObject and self.lastObjectCounter <= 5:
				self.lastObjectCounter += 1
				oName = self.lastObject[0]
				user = self.lastObject[1]
				oPos = self.lastObject[2]
				oRot = self.lastObject[3]
				p = bf2.PlayerManager.Player(user)
				pD = self.pdata[p.getName()]
				if type(oPos) == tuple and type(oRot) == tuple and p.isAlive():
					objectsOfTemplate = bf2.objectManager.getObjectsOfTemplate(oName)
					print len(objectsOfTemplate)
					o = None
					o = objectsOfTemplate[-1]
					print o
					if o.isValid() and not self.preCompletedObjects.count(o):
						pass
					else:
						o = None
					if o:
						print o.getPosition()
						o.setPosition(oPos)
						self.preCompletedObjects.append(o)
						#if self.is_spGame:
						#	objs = host.rcon_invoke("object.listObjectsofTemplate " + str(oName))
						#	objs = objs.split()
						#	o.oId = objs[-7]
						pD.objects.append(o)
						if pD.autoGrab and ((not self.allowed_vehicles.count(oName)) or (self.autograb_excempt.count(oName))):
							if pD.grabMode <= 2:
								pD.grabbed = [o,pD.tracerDistance,1,o.getPosition(),o.getRotation()]
							else:
								pPos = p.getDefaultVehicle().getPosition()
								pRot = p.getDefaultVehicle().getRotation()
								pCam = self.getSoldierCam(p)
								pCamRot = pCam.getRotation()
								sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
								shp = o.getPosition()
								grabOffset = ((sPos[0] - shp[0]),(sPos[1] - shp[1]),(sPos[2] - shp[2]))
								pD.grabbed = [o,pD.tracerDistance,1,o.getPosition(),o.getRotation(),grabOffset]
						self.lastObject = []
						self.lastObjectCounter = 0
		except:
			ExceptionOutput()

		for x in bf2.playerManager.getPlayers():
			try:
				p = bf2.PlayerManager.Player(x.index)
				pD = self.pdata[p.getName()]
				if pD.spawnDelay > 0:
					pD.spawnDelay -= 1
				if p.isAlive():
					pPos = p.getDefaultVehicle().getPosition()
					pRot = p.getDefaultVehicle().getRotation()
					#### Get all grabbed objects and move them where they need to be
					if pD.grabbed:
						if pD.grabbed[0].isValid():
							pCam = self.getSoldierCam(p)
							pCamRot = pCam.getRotation()
							sPos = self.getVectoredPosition(pPos,pD.grabbed[1],pRot,pCamRot)
							if pD.gridSnap:
								sPos = (math.ceil(sPos[0] / pD.gridSnap) * pD.gridSnap,math.ceil(sPos[1] / pD.gridSnap) * pD.gridSnap,math.ceil(sPos[2] / pD.gridSnap) * pD.gridSnap)
							if pD.grabbed[0].getPosition() != sPos:
								if pD.grabMode <= 2:
									pD.grabbed[0].setPosition(sPos)
								else:
									pD.grabbed[0].setPosition(((sPos[0] - pD.grabbed[5][0]),(sPos[1] - pD.grabbed[5][1]),(sPos[2] - pD.grabbed[5][2])))
								#self.dbg("new position: " + str(sPos[0]) + "," + str(sPos[1]) + "," + str(sPos[2]))
							pD.grabbed[0].setRotation(pD.grabbed[4])
						else:
							pD.grabbed = []
					if pD.preGrabMode == 1 and self.timerCounter == 0:
						pCam = self.getSoldierCam(p)
						shortest = []
						shortest = [0,pD.tracerDistance + 10.0]
						pCamRot = pCam.getRotation()
						sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
						sRot = pRot
						for v in pD.objectTypes:
							if self.allowed_objects.count(v):
								for q in bf2.objectManager.getObjectsOfTemplate(v):
									if q.isValid():
										qDist = self.getVectorDistance(sPos,q.getPosition())
										if qDist < shortest[1]:
											shortest = [q,qDist]
						if pD.preGrab:
							if shortest[1] > pD.preGrab[1]:
								shortest[0] = pD.preGrab[0]
								shortest[1] = pD.preGrab[1]
						else: pD.preGrab = [shortest[0],shortest[1],1,shortest[0].getPosition(),pCam.getRotation()]
						if pD.preGrab[0] != shortest[0]:
							# Object, current flash thing, object position, player rotation
							o = pD.preGrab[0]
							if o.isValid():
								if hasattr(o,"oId"):
									host.rcon_invoke("Object.active id" + str(o.oId))
									host.rcon_invoke("Object.isVisible 1")
								else: o.setPosition(pD.preGrab[3])
								pD.preGrab[1] = 1
							pD.preGrab = [shortest[0],shortest[1],1,shortest[0].getPosition(),pCam.getRotation()]
						else:
							if pD.preGrab[1] == 1:
								o = pD.preGrab[0]
								if o.isValid():
									if hasattr(o,"oId"):
										host.rcon_invoke("Object.active id" + str(o.oId))
										host.rcon_invoke("Object.isVisible 0")
									else: o.setPosition((0.0,83453.0,0.0))
								pD.preGrab[1] = 2
							else:
								o = pD.preGrab[0]
								if o.isValid():
									if hasattr(o,"oId"):
										host.rcon_invoke("Object.active id" + str(o.oId))
										host.rcon_invoke("Object.isVisible 1")
									else: o.setPosition(pD.preGrab[3])
								pD.preGrab[1] = 1
			except:
				pass

	def freezer(self,data):
		#### Refreeze all frozen objects
		invalidFreezes = []
		for o in self.obj_freeze:
			if self.obj_freeze[o]:
				if (o.isValid()) and (type(self.obj_freeze[o][0]) == type(('tuple','thing'))):
					if o.getPosition() != self.obj_freeze[o][0]:
						o.setPosition(self.obj_freeze[o][0])
					if o.getRotation() != self.obj_freeze[o][1]:
						o.setRotation(self.obj_freeze[o][1])
				else:
					invalidFreezes.append(o)
		for f in invalidFreezes:
			if self.obj_freeze.has_key(f):
				self.obj_freeze.pop(f)

	#######################################
	#      Sandbox utility functions      #
	#######################################
	#######################################

	def createSpawnerObject(self, sName, sPos, sRot = (0.0,0.0,0.0), user=0):
		if not self.spawnerTemplates.count(sName):
			self.createSpawnerTempl(sName, sName)
			self.spawnerTemplates.append(sName)
		print host.rcon_invoke("Object.create sbx_" + sName)
		print host.rcon_invoke("Object.absolutePosition " + str(sPos[0]) + "/" + str(sPos[1]) + "/" + str(sPos[2]))
		print host.rcon_invoke("Object.rotation " + str(sRot[0]) + "/" + str(sRot[1]) + "/" + str(sRot[2]))
		print host.rcon_invoke("Object.setControlPointId 506")
		if sName == "usaas_stinger_spw" and user:
			p = bf2.PlayerManager.Player(user)
			host.rcon_invoke("Object.team " + str(p.getTeam()))
			host.rcon_invoke("ObjectTemplate.teamOnVehicle 1")
		elif sName == "usaas_stinger_spw":
			host.rcon_invoke("Object.team 2")
			host.rcon_invoke("ObjectTemplate.teamOnVehicle 1")
		else:
			host.rcon_invoke("Object.team 2")
		#host.rcon_invoke("Object.delete")
		if user and not (self.allowed_vehicles.count(sName) and self.autograb_excempt.count(sName)):
			self.lastObject = [str(sName),user,sPos,sRot]
			self.lastObjectCounter = 0

	def createSpawnerTempl(self, id, tmpl):
		host.rcon_invoke("ObjectTemplate.create ObjectSpawner sbx_" + id)
		host.rcon_invoke("ObjectTemplate.activeSafe ObjectSpawner sbx_" + id)
		host.rcon_invoke("ObjectTemplate.hasMobilePhysics 0")
		host.rcon_invoke("ObjectTemplate.setObjectTemplate 1 " + tmpl)
		host.rcon_invoke("ObjectTemplate.setObjectTemplate 2 " + tmpl)
		host.rcon_invoke("ObjectTemplate.minSpawnDelay 700000")
		host.rcon_invoke("ObjectTemplate.maxSpawnDelay 900000")

	def createSpawner(self, id, sPos, sRot):
		host.rcon_invoke("Object.create sbx_" + id)
		host.rcon_invoke("Object.absolutePosition " + str(sPos[0]) + "/" + str(sPos[1]) + "/" + str(sPos[2]))
		host.rcon_invoke("Object.rotation " + str(sRot[0]) + "/" + str(sRot[1]) + "/" + str(sRot[2]))
		host.rcon_invoke("Object.team 2")
		host.rcon_invoke("Object.delete")

	def createObject(self, oName, oPos, oRot = (0.0,0.0,0.0), user = 0):
		try:
			oPos = (float(oPos[0]),float(oPos[1]),float(oPos[2]))
			oRot = (float(oRot[0]),float(oRot[1]),float(oRot[2]))
			host.rcon_invoke("gameLogic.createObject " + str(oName))
			self.lastObject = [str(oName),user,oPos,oRot]
			self.lastObjectCounter = 0
		except:
			ExceptionOutput()

	def deleteObject(self, o):
		try:
			if o:
				if o.isValid():
					if hasattr(o,"oId"):
						host.rcon_invoke("Object.active id" + str(o.oId))
						host.rcon_invoke("Object.delete")
					else:
						o.setPosition((0.0,99999.0,0.0))
						if o.isValid() and o.templateName == "usaas_stinger_spw":
							try:
								o.setDamage(1)
							except:
								pass
						self.deleted_objects.append(o)
		except:
			ExceptionOutput()

	def sendMsg(self, p, msg):
		if p:
			host.rcon_invoke("game.sayAll \"" + str(p.getName()) + " - " + str(msg) + "\"")
		else:
			host.rcon_invoke("game.sayAll \"" + str(msg) + "\"")

	def getSoldierCam(self, player):
		p = player.getDefaultVehicle()
		kids = p.getChildren()
		for child in kids:
			if child.templateName == "SoldierCamera":
				return child

	def checkAdmin(self, pName):
		if self.is_spGame:
			return True
		try:
			pName = pName.split(" ")[-1]
			if self.admins.lower().count(pName.lower()):
				return True
			else:
				return False
		except:
			return False

	#######################################
	#         Saving and Loading          #
	#######################################
	#######################################

	def saveSandbox(self, filename):
		file_string = str(host.sgl_getMapName()) + "," + str(self.SANDBOX_VERSION) + "@"
		out_file = open("mods/sandbox/saved/" + str(filename) + ".bf2sbx","w")
		increment = 0
		for t in self.object_types:
			for q in bf2.objectManager.getObjectsOfTemplate(t):
				if q.isValid():
					if not self.deleted_objects.count(q):
						increment += 1
						qPos = q.getPosition()
						qRot = q.getRotation()
						qNam = q.templateName
						file_string += str(qNam) + ";" + str(qPos[0]) + "," + str(qPos[1]) + "," + str(qPos[2]) + ";" + str(qRot[0]) + "," + str(qRot[1]) + "," + str(qRot[2]) + "!"
		out_file.write(file_string)
		out_file.close()
		self.sendMsg(0, "Your sandbox has been saved as " + str(filename) + ".bf2sbx")
		self.dbg("Saved file, " + str(increment) + " objects")
		if self.ftpUpload and not self.is_spGame:
			result = sandboxUploader.sUpload("mods/sandbox/saved/" + str(filename) + ".bf2sbx",self.ftpOptions)
			if result:
				self.sendMsg(0, "The sandbox has also been uploaded to " + str(self.ftpOptions['dlLocation']))

	def loadSandbox(self, filename, p):
		pD = self.pdata[p.getName()]
		try: in_file = open("mods/sandbox/saved/" + str(filename) + ".bf2sbx","r")
		except:
			self.sendMsg(0, "The file you specified was not found.")
			return 0
		savedStr = in_file.read()
		a = savedStr.split("@")
		w = a[0].split(",")
		if not w.count(self.SANDBOX_VERSION):
			self.sendMsg(0, "This file was created with a different version of BF2:Sandbox. Some objects may be loaded incorrectly.")
		self.sendMsg(0, "Loading sandbox file...this may lag for a little while.")
		if w[0].lower() == str(host.sgl_getMapName()).lower():
			b = a[1].split("!")
			increment = 0
			for objInfo in b:
				if objInfo:
					objInfoList = objInfo.split(";")
					oName = objInfoList[0]
					oPos = objInfoList[1].split(",")
					oRot = objInfoList[2].split(",")
					if self.mp_objects.count(oName.lower()) and not self.is_spGame:
						oName += "_mp"
					self.createSpawnerObject(oName, oPos, oRot)
					if self.object_types.count(oName) < 1:
						self.object_types.append(oName)
					increment += 1
			pD.objectTypes = self.object_types
			self.sendMsg(0, str(filename) + ".bf2sbx has been loaded.")
			self.dbg("Loaded file, " + str(increment) + " objects")
		else:
			self.sendMsg(0, "The file you specified was not created in this map.")
		in_file.close()

	def saveCon(self, saveVehicles=False):
		increment = 0
		object_str = ""
		vehicle_str = ""
		vehicle_spawner_str = ""
		for t in self.object_types:
			for q in bf2.objectManager.getObjectsOfTemplate(t):
				if q.isValid():
					if not self.deleted_objects.count(q):
						increment += 1
						qPos = q.getPosition()
						qRot = q.getRotation()
						qNam = q.templateName
						if not q.isPlayerControlObject:
							object_str += "\nrem *** " + str(qNam) + " ***"
							object_str += "\nObject.create " + str(qNam)
							object_str += "\nObject.absolutePosition " + str(qPos[0]) + "/" + str(qPos[1]) + "/" + str(qPos[2])
							object_str += "\nObject.rotation " + str(qRot[0]) + "/" + str(qRot[1]) + "/" + str(qRot[2])
							object_str += "\nObject.layer 1\n"
						elif q.isPlayerControlObject and saveVehicles:
							vehicle_str += "\nrem [ObjectSpawnerTemplate: sandbox" + str(increment) + "]"
							vehicle_str += "\nObjectTemplate.create ObjectSpawner sandbox" + str(increment)
							vehicle_str += "\nObjectTemplate.activeSafe ObjectSpawner sandbox" + str(increment)
							vehicle_str += "\nObjectTemplate.modifiedByUser elxxrox"
							vehicle_str += "\nObjectTemplate.isNotSaveable 1"
							vehicle_str += "\nObjectTemplate.hasMobilePhysics 0"
							vehicle_str += "\nObjectTemplate.setObjectTemplate 1 " + str(qNam)
							vehicle_str += "\nObjectTemplate.setObjectTemplate 2 " + str(qNam)
							vehicle_str += "\n"
							vehicle_spawner_str += "\n   rem [ObjectSpawner: sandbox" + str(increment) + "]"
							vehicle_spawner_str += "\n   Object.create sandbox" + str(increment)
							vehicle_spawner_str += "\n   Object.absolutePosition " + str(qPos[0]) + "/" + str(qPos[1]) + "/" + str(qPos[2])
							vehicle_spawner_str += "\n   Object.rotation " + str(qRot[0]) + "/" + str(qRot[1]) + "/" + str(qRot[2])
							vehicle_spawner_str += "\n   Object.setControlPointId " + self.con_save_cp
							vehicle_spawner_str += "\n   Object.layer 1\n"
		file = open("mods/sandbox/saved/staticobjects_" + str(host.sgl_getMapName()).lower() + ".txt","w")
		file.write(object_str)
		file.close()
		if saveVehicles:
			file = open("mods/sandbox/saved/gameplayobjectsTemplates_" + str(host.sgl_getMapName()).lower() + ".txt","w")
			file.write(vehicle_str)
			file.close()
			file = open("mods/sandbox/saved/gameplayobjectsSpawners_" + str(host.sgl_getMapName()).lower() + ".txt","w")
			file.write(vehicle_spawner_str)
			file.close()
		self.sendMsg(0, "Saved files for map creation.")

	def dbg(self,msg):
		if self.enableDebug:
			print msg

	#######################################
	#      Battlefield engine events      #
	#######################################
	#######################################

	def onChatMessage(self, playerId, text, channel, flags):
		text = text.replace( "HUD_TEXT_CHAT_TEAM", "" )
		text = text.replace( "HUD_TEXT_CHAT_SQUAD", "" )
		text = text.replace( "HUD_CHAT_DEADPREFIX", "" )
		text = text.replace( "*\xA71DEAD\xA70*", "" )
		if text[0:1] == "@":
			pattern = re.compile(r'@(\w*) ?(.*)')
			matches = pattern.findall(text)
			command = matches[0][0]
			if matches[0][1] != "":
				parameter = matches[0][1]
			else:
				parameter = None
			if playerId == -1: #fix for local servers
				playerId = 255
			p = bf2.PlayerManager.Player(playerId)
			pD = self.pdata[p.getName()]
			if self.checkAdmin(p.getName()) or self.is_spGame == 1:
				if command == "save" and parameter:
					self.saveSandbox(parameter)
				if command == "savecon" and parameter:
					self.saveCon(True)
				elif command == "savecon":
					self.saveCon()
				if command == "load" and parameter:
					self.loadSandbox(parameter,p)
				if command == "r" and parameter:
					self.onRemoteCommand(playerId,parameter)
				if command == "s" and parameter:
					self.onRemoteCommand(playerId,"btnSpawn " + parameter)
				if command == "clear":
					host.rcon_invoke("admin.restartmap")

	def onPlayerSpawn(self, player, soldier):
		if self.is_spGame:
			if player.isRemote():
				self.is_spGame = 0

	def onEnterVehicle(self, player, vehicle, freeSoldier = False):
		pD = self.pdata[player.getName()]
		pD.grabbed = []

	def onExitVehicle(self, player, vehicle):
		pD = self.pdata[player.getName()]
		pD.grabbed = []

	def onPlayerDeathCQ(self, victim, vehicle):
		pD = self.pdata[victim.getName()]
		pD.grabbed = []

	def onPlayerKilledCQ(self, victim, attacker, weapon, assists, object):
		pD = self.pdata[victim.getName()]
		pD.grabbed = []

	def onPlayerDisconnect(self, player):
		if self.pdata.has_key(player.getName()):
			pD = pdata[player.getName()]
			pD.grabbed = []
			for o in pD.objects:
				self.deleteObject(o)
		for o in self.obj_locked:
			if self.obj_locked[o] == player.getName():
				self.obj_locked.pop(o)
		try:
			self.pdata.pop(player.getName())
		except:
			ExceptionOutput()

	#######################################
	#        Mathematics functions        #
	#######################################
	#######################################

	def getVectoredPosition(self, pPos, distance, pRot, pCamRot):
		from math import pi
		#Calculate the height using a triangle
		p = pCamRot[1] #Pitch of the player's cam
		a = 90.0 - abs(p)
		p = pi * p / 180.0
		a = pi * a / 180.0
		sVz = (distance * math.sin(p))/math.sin(a)
		#convert from crazy-ass BF2 values to polar coordinate values
		new_pRot = -(pRot[0] + pCamRot[0]) + 90.0
		sVx,sVy = self.rect(distance,new_pRot,1) #Convert vector of [spawnDistance] with heading of cam rotation to x,y values
		sPos = (pPos[0] + sVx, pPos[1] - sVz, pPos[2] + sVy)
		return sPos

	def rect(self, r, w, deg=0): # radian if deg=0; degree if deg=1
		from math import cos, sin, pi
		if deg:
			w = pi * w / 180.0
		return r * cos(w), r * sin(w)

	def getVectorDistance(self, pos1, pos2):
		diffVec = [0.0, 0.0, 0.0]
		diffVec[0] = math.fabs(pos1[0] - pos2[0])
		diffVec[1] = math.fabs(pos1[1] - pos2[1])
		diffVec[2] = math.fabs(pos1[2] - pos2[2])
		return math.sqrt(diffVec[0] * diffVec[0] + diffVec[1] * diffVec[1] + diffVec[2] * diffVec[2])

	#######################################
	# Utility functions for other modders #
	#######################################
	#######################################

	def disableBoundaries(self):
		host.rcon_invoke("CombatAreaManager.use 0")
		host.rcon_invoke("CombatAreaManager.timeAllowedOutside 1000.000000")
		self.sBoundaries = False

	def enableBoundaries(self):
		host.rcon_invoke("CombatAreaManager.use 1")
		host.rcon_invoke("CombatAreaManager.timeAllowedOutside 10.00")
		self.sBoundaries = True

	def setAllowedObjects(self,objList):
		if type(self.allowed_objects) == type(objList):
			self.allowed_objects = objList

	def addAllowedObjects(self,obj):
		# This can take either a list or a string.
		if type(self.allowed_objects) == type(obj):
			self.allowed_objects.extend(obj)
		else:
			self.allowed_objects.append(obj)

	def delAllowedObjects(self,obj):
		# This can take either a list or a string.
		if type(self.allowed_objects) == type(obj):
			for x in obj:
				try:
					self.allowed_objects.remove(x)
				except:
					pass #do nothing
		else:
			self.allowed_objects.remove(obj)

	def setAllowLock(self,val):
		if val: self.allow_lock = True
		else: self.allow_lock = False

	def setAllowGrab(self,val):
		if val: self.allow_grab = True
		else: self.allow_grab = False

	def setAllowCreate(self,val,override=True):
		if val: self.allow_create = True
		else: self.allow_create = False
		if override: self.allowCreateOverride = True
		else: self.allowCreateOverride = False

	def setMinPlayers(self,val):
		host.rcon_invoke("sv.numPlayersNeededToStart " + str(val))

def start():
	return sandboxCore()

print "Sandbox core loaded."