#*************************************************************************************************************
# Name: SandboxCore
# Author: Elxx (http://elxx.net)
# Version 0.4 ALPHA, BF2142 Optimized
#*************************************************************************************************************
# Conditions and Limitations: You may use this file locally, on listening and dedicated servers.
# This software is provided "as is". In no event shall the author, authors or contributors be held
# liable for  any direct, indirect, incidental, special, exemplary, or consequential damages (including)
# but not limited to, procurement of substitute good or services; loss of use, data or profits; or
# business interruption) however caused and on any theory of liability, whether in contract, strict liability
# arising in any way out of the use of this software, even if advised of the possibility of such damage.
# This file may be modified as long as this header remains intact. The BF2:Sandbox team offers no support
# for modified versions of this file.
#*************************************************************************************************************
#
# Configure the users who can use the Admin Tools in Multiplayer.
MP_LS_USERS = ['SampleUser1','Sir SampleAdmin2']
#
#*************************************************************************************************************

import host
import bf2
import bf2.Timer
import bf2.PlayerManager
import math
import sys
import re
import os
import inspect
from bf2.stats.constants import *

class pData:
	def __init__(self):
		self.ObjectRestrictions = False
		self.tracerDistance = 3
		self.objects = []
		self.objectTypes = []
		self.grabbed = []
		self.lastRotation = 0
		self.lastSpawned = []
		self.lastTemplate = ""
		self.tracerOn = 1
		self.tracerDistance = 8
		self.gridSnap = 3
		self.rotSnap = 5
		self.autoGrab = 1
		self.numSpawned = [0,0]
		self.preGrab = []
		self.preGrabMode = 0
		self.lastAction = []
		self.spawnDelay = 0
		self.projectileGrab = 0

# Actions
# 1 = grab [object, original position, original rotation]
# 2 = drop [object, frozen]
# 3 = delete [object, original position]
# 4 =

def ExceptionOutput():
	sys.stderr.write("\n" + "Exception Occured: " + str(sys.exc_info()[0]) + "\n")
	sys.stderr.write("Value: " + str(sys.exc_info()[1]) + "\n")
	sys.stderr.write("Line:" + str(readline(inspect.getfile(sys.exc_info()[2]),
							sys.exc_info()[2].tb_lineno)) + "\n")
	sys.stderr.write("Line #: " + str(sys.exc_info()[2].tb_lineno) + "\n")
	sys.stderr.write("File: " + str(inspect.getfile(sys.exc_info()[2])) + "\n" + "\n")
	#host.rcon_invoke("game.sayAll \"Exception Caught: " + str(sys.exc_info()[1]) + "\"")
	#host.rcon_invoke("game.sayAll \"Line " + str(sys.exc_info()[2].tb_lineno) + "\"")

def readline(filename, lineno):
	filen = re.sub('\\\\', '/', filename)
	file = open(filen, 'rU')
	lines = file.readlines()
	file.close()
	linen = lineno - 1
	line = re.sub('\s+', ' ', lines[linen])
	return line

SANDBOX_VERSION = "0.4"

class sandboxCore:
	def __init__(self):
		self.SANDBOX_VERSION = SANDBOX_VERSION
		# Set up event handlers
		host.registerGameStatusHandler(self.onGameStatusChanged)
		host.registerHandler('PlayerDeath', self.onPlayerDeathCQ)
		host.registerHandler('PlayerKilled', self.onPlayerKilledCQ)
		host.registerHandler('PlayerSpawn', self.onPlayerSpawn)
		host.registerHandler('EnterVehicle', self.onEnterVehicle)
		host.registerHandler('ExitVehicle', self.onExitVehicle)
		host.registerHandler('PlayerDisconnect', self.onPlayerDisconnect)
		host.registerHandler('ChatMessage', self.onChatMessage, 1)
		host.registerHandler('RemoteCommand', self.onRemoteCommand, 1)
		self.SPAWN_LIMIT = 200
		self.RAGDOLL_LIMIT = 15
		self.ObjectRestrictions = False

	def initSbx(self):
		# Set up class variables
		self.object_types = []
		self.deleted_objects = []
		self.obj_freeze = {}
		self.obj_locked = {}
		self.is_spGame = 0
		self.ragdoll_objects = []
		self.annoying_objects = []
		self.mp_objects = ['container_empty_c','container_seald_d']
		self.nonmp_objects = []
		self.allowed_objects = ['container_empty_c','container_seald_d','fence_industry_16m','fence_corrugated_8m','crate01_c','barrel_big_c']
		self.allowed_vehicles = ['eu_fav','as_fav','us_heavy_mech','as_heavy_mech','eu_ag','as_ag']
		self.sbxTimer = None
		self.freezerTimer = None
		self.SPAWN_LIMIT = 2000
		self.RAGDOLL_LIMIT = 15
		self.p_objects = {}
		self.MP_LS_USERS = MP_LS_USERS
		self.allow_lock = True
		self.allow_grab = True
		self.allow_create = True
		self.gravity = -15
		self.timerLoaderActive = 0
		self.timerLoader = [1,0,""]
		self.timerLoaderObjects = []
		self.timerLoaderTempFreeze = {}
		self.spawnerTemplates = []
		self.lastObject = []
		self.allowCreateOverride = True
		self.sBoundaries = True
		self.pdata = {}
		self.preCompletedObjects = []
		self.calculatedProjectiles = []
		# Set up the sandbox timer
		self.timerCounter = 0
		if not self.sbxTimer:
			self.sbxTimer = bf2.Timer(self.run_sbxTimer, 0 , 1)
			self.sbxTimer.setRecurring(0.01)
		if not self.freezerTimer:
			self.freezerTimer = bf2.Timer(self.freezer, 0 , 1)
			self.freezerTimer.setRecurring(0.01)
		# Debug mode?
		self.enableDebug = 1
		self.dbg("Sandbox has been initialized.")

	def dbg(self,msg):
		if self.enableDebug:
			print msg

	def onGameStatusChanged(self, status):
		if status == bf2.GameStatus.Playing:
			self.preCompletedObjects = []
			#for x in bf2.objectManager.getObjectsOfType('dice.hfe.world.ObjectTemplate.PlayerControlObject'):
			#	self.preCompletedObjects.append(x)
			self.p_objects = {}
			self.pdata = {}
			self.object_types = []
			self.deleted_objects = []
			self.obj_freeze = {}
			self.obj_locked = {}
			self.gravity = -15
			self.timerLoaderActive = 0
			self.timerLoader = [1,0,""]
			self.timerLoaderObjects = []
			self.timerLoaderTempFreeze = {}
			self.spawnerTemplates = []
			host.rcon_invoke("physics.gravity " + str(self.gravity))
			if self.sBoundaries:
				self.enableBoundaries()
			else:
				self.disableBoundaries()
			self.dbg("Game status cleared and Sandbox reloaded.")

	def onRemoteCommand(self, playerId, cmd):
		from math import pi

		if playerId == -1: #fix for local servers
			playerId = 255
		self.dbg("Command received: [" + str(cmd) + "] from [" + str(playerId) + "]")
		p = bf2.PlayerManager.Player(playerId)
		pD = self.pdata[p.getName()]
		pPos = p.getDefaultVehicle().getPosition()
		pRot = p.getDefaultVehicle().getRotation()
		origCmd = cmd
		spacepos = origCmd.find(' ')
		if spacepos == -1: spacepos=len(cmd)
		cmd = origCmd[0:spacepos]
		cmdVar = origCmd[spacepos + 1:]
		if cmd == "btnSpawn":
			if self.ObjectRestrictions:
				if not self.allow_create and (not self.MP_LS_USERS.count(p.getName()) and not self.allowCreateOverride):
					return 0
				if cmdVar and (not self.allowed_objects.count(cmdVar.lower()) and not self.allowed_vehicles.count(cmdVar.lower())):
					return 0
				if pD.spawnDelay > 0:
					return 0
				if self.nonmp_objects.count(cmdVar.lower()) and not self.is_spGame:
					return 0
			if self.mp_objects.count(cmdVar.lower()) and not self.is_spGame:
			#if self.mp_objects.count(cmdVar.lower()):
				cmdVar += "_mp"
			pCam = self.getSoldierCam(p)
			pCamRot = pCam.getRotation()
			sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
			sRot = pRot
			self.dbg("new position: " + str(sPos[0]) + "," + str(sPos[1]) + "," + str(sPos[2]))
			self.dbg("new rotation: " + str(sRot[0]) + "," + str(sRot[1]) + "," + str(sRot[2]))
			if cmdVar:
				pD.lastTemplate = str(cmdVar)
				for q in bf2.objectManager.getObjectsOfTemplate(cmdVar):
					if not self.preCompletedObjects.count(q):
						self.preCompletedObjects.append(q)
			else:
				if pD.lastRotation:
					sRot = pD.lastRotation
			cmdVar = pD.lastTemplate
			# Create the spawner using the fancy new method
			#o = self.createObject(str(cmdVar), sPos, sRot, p.getName())
			try:
				#self.createSpawnerObject(str(cmdVar), sPos, sRot, p.index)
				self.createObject(str(cmdVar), sPos, sRot, p.index)
			except:
				ExceptionOutput()
			if pD.objectTypes.count(str(cmdVar)) < 1:
				pD.objectTypes.append(str(cmdVar))
			if self.object_types.count(str(cmdVar)) < 1:
				self.object_types.append(str(cmdVar))
			pD.lastRotation = sRot
			if not self.is_spGame:
				if self.annoying_objects.count(str(cmdVar)):
					pD.spawnDelay = 120
				else:
					pD.spawnDelay = 80
			self.dbg("Created spawner!")

		if cmd == "btnGrabAlt":
			if self.is_spGame:
				if pD.preGrabMode == 0:
					pD.preGrabMode = 1
				else:
					o = pD.preGrab[0]
					if hasattr(o,"oId"):
						host.rcon_invoke("Object.active id" + str(o.oId))
						host.rcon_invoke("Object.isVisible 1")
					pD.preGrab[0].setPosition(pD.preGrab[3])
					pD.preGrabMode = 0
					pD.preGrab = []

		if cmd == "btnGrab":
			if pD.spawnDelay > 40:
				return 0
			if not self.allow_grab and not self.MP_LS_USERS.count(p.getName()):
				return 0
			if pD.preGrabMode == 1 and self.is_spGame:
				o = pD.preGrab[0]
				if hasattr(o,"oId"):
					host.rcon_invoke("Object.active id" + str(o.oId))
					host.rcon_invoke("Object.isVisible 1")
				pD.preGrab[0].setPosition(pD.preGrab[3])
				pD.preGrabMode = 0
				pD.preGrab = []
			shortest = [0,pD.tracerDistance + 20.0]
			pCam = self.getSoldierCam(p)
			pCamRot = pCam.getRotation()
			if pD.projectileGrab:
				sPos = pD.projectileGrab
				pD.projectileGrab = 0
			else:
				sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
			sRot = pRot
			#for v in self.object_types:
			for v in pD.objectTypes:
				for q in bf2.objectManager.getObjectsOfTemplate(v):
					if q.isValid():
						qDist = self.getVectorDistance(sPos,q.getPosition())
						if qDist < shortest[1]:
							shortest = [q,qDist]
			if shortest[0]:
				if not self.is_spGame and self.obj_locked.has_key(shortest[0]):
					if self.obj_locked[shortest[0]] != p.index: return 0
				self.dbg("Found closest object")
				distance = self.getVectorDistance(pPos,shortest[0].getPosition())
				frozen = 0
				if self.obj_freeze.has_key(shortest[0]):
					frozen = 1
					self.obj_freeze.pop(shortest[0])
				pD.grabbed = [shortest[0],distance,1,shortest[0].getPosition(),shortest[0].getRotation()]
				pD.lastAction.append([1,shortest[0],shortest[0].getPosition(),shortest[0].getRotation(),frozen])
				# 1 = grab [object, original position, original rotation, original frozen]
				self.dbg("Grabbed closest object")
				pD.spawnDelay = 80

		if cmd == "btnRotate":
			v = cmdVar.split(" ")
			if pD.grabbed:
				if v[1] == "1": rFactor = pD.rotSnap
				else: rFactor = -pD.rotSnap
				obj = pD.grabbed[0]
				if obj and obj.isValid():
					objRot = obj.getRotation()
					if v[0] == "A":
						newRot = (objRot[0] + rFactor, objRot[1], objRot[2])
					if v[0] == "P":
						newRot = (objRot[0], objRot[1] + rFactor, objRot[2])
					if v[0] == "R":
						newRot = (objRot[0], objRot[1], objRot[2] + rFactor)
					pD.grabbed[4] = newRot
					obj.setRotation(newRot)
					pD.lastRotation = newRot

		if cmd == "btnDrop":
			pD.lastAction.append([2,pD.grabbed[0],pD.grabbed[1],0])
			# 2 = drop [object, distance]
			pD.grabbed = []

		if cmd == "btnFreeze":
			if pD.grabbed:
				if pD.grabbed[0].isValid():
					self.obj_freeze[pD.grabbed[0]] = [pD.grabbed[0].getPosition(),pD.grabbed[0].getRotation()]
					pD.lastAction.append([2,pD.grabbed[0],pD.grabbed[1],1])
					pD.grabbed = []

		if cmd == "btnLock":
			if pD.grabbed and self.allow_lock:
				if pD.grabbed[0].isValid():
					self.obj_locked[pD.grabbed[0]] = p.index

		if cmd == "btnUnLock":
			if pD.grabbed:
				if pD.grabbed[0].isValid():
					try:
						del self.obj_locked[pD.grabbed[0]]
					except:
						pass #do nothing

		if cmd == "btnDelete":
			if pD.grabbed:
				if pD.grabbed[0].isValid():
					o = pD.grabbed[0]
					pD.grabbed = []
					self.deleteObject(o)

		if cmd == "btnPush":
			if cmdVar == "1":
				pD.grabbed[1] += 0.5
			else:
				pD.grabbed[1] += 1.0

		if cmd == "btnPull":
			if cmdVar == "1":
				pD.grabbed[1] -= 0.5
			else:
				pD.grabbed[1] -= 1.0

		if cmd == "btnUndo":
			#if pD.lastAction and self.is_spGame:
			if pD.lastAction:
				a = pD.lastAction.pop()
				if a[0] == 1: #grab [object, original position, original rotation, original frozen]
					self.dbg("Grab undoing")
					o = a[1]
					if pD.grabbed[0] == o:
						pD.grabbed = []
					o.setPosition(a[2])
					o.setRotation(a[3])
					if a[4] or (not self.is_spGame):
						self.obj_freeze[o] = [a[2],a[3]]
						self.dbg("Refrozen")
				if a[0] == 2: # 2 = drop [object, distance]
					self.dbg("Drop undoing")
					o = a[1]
					pD.grabbed = []
					pD.grabbed = [o,a[2],1,o.getPosition(),o.getRotation()]
				if a[0] == 3: # 3 = delete [object name, original position, original rotation]
					self.dbg("Delete undoing")

					#o = a[1]
					#o.setPosition(a[2])
					#self.deleted_objects.remove(o)
					#pD.numSpawned[0] += 1
					#self.onRemoteCommand(playerId,"btnUndo") #Undo once more so we re-grab the deleted object
				self.dbg("Undone.")

		if cmd == "btnTracerToggle":
			'''children = p.getPrimaryWeapon().getChildren()
			for child in children:
				if child.templateName == "glowshit":
					pos = child.getPosition()
					if pD.tracerOn:
						child.setPosition((pos[0],pos[1] - 100.0,pos[2]))
						pD.tracerOn = 0
					else:
						child.setPosition((pos[0],pos[1] + 100.0,pos[2]))
						pD.tracerOn = 1'''

		if cmd == "btnAutoGrabToggle":
			if pD.autoGrab:
				pD.autoGrab = 0
				self.sendMsg(p,"Autograb disabled.")
			else:
				pD.autoGrab = 1
				self.sendMsg(p,"Autograb enabled.")

		if cmd == "btnTracerDist":
			if cmdVar == "push":
				if pD.tracerDistance >= 15:
					pD.tracerDistance += 5
				else:
					pD.tracerDistance += 1
			else:
				if pD.tracerDistance > 15:
					pD.tracerDistance -= 5
				else:
					pD.tracerDistance -= 1
			self.sendMsg(p, "Tracer distance set to " + str(pD.tracerDistance))
			'''children = p.getPrimaryWeapon().getChildren()
			for child in children:
				if child.templateName == "glowshit":
					pPos = p.getPrimaryWeapon().getPosition()
					pCam = self.getSoldierCam(p)
					pCamRot = pCam.getRotation()
					sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
					child.setPosition(sPos)'''

		if cmd == "btnRotSnap":
			if cmdVar == "inc":
				if pD.rotSnap > 4:
					pD.rotSnap += 5
				else:
					pD.rotSnap += 1
			else:
				if pD.rotSnap > 5:
					pD.rotSnap -= 5
				elif pD.rotSnap <= 5 and pD.rotSnap > 1:
					pD.rotSnap -= 1
			self.sendMsg(p, "Rotation factor set to " + str(pD.rotSnap) + "deg")

		if cmd == "btnSnap":
			if cmdVar == "inc":
				if pD.gridSnap > 0:
					pD.gridSnap -= 1
				else:
					self.sendMsg(p, "Grid snap is already at maximum level.")
			else:
				if pD.gridSnap < 3:
					pD.gridSnap += 1
			if pD.gridSnap < 3:
				self.sendMsg(p, "Grid snap set to " + str(pD.gridSnap))
			else:
				self.sendMsg(p, "Grid snap disabled")

		if self.MP_LS_USERS.count(p.getName()) or self.is_spGame == 1:

			if cmd == "btnGravity" and cmdVar:
				try:
					gravFactor = int(cmdVar)
					self.gravity += gravFactor
					self.sendMsg(0,"Gravity set to " + str(self.gravity))
					host.rcon_invoke("physics.gravity " + str(self.gravity))
				except:
					ExceptionOutput()

			if cmd == "btnAdGrab" and cmdVar:
				if cmdVar == "1":
					self.allow_grab = True
					self.sendMsg(0,"Object grab enabled")
				else:
					self.allow_grab = False
					self.sendMsg(0,"Object grab disabled")

			if cmd == "btnAdLock" and cmdVar:
				if cmdVar == "1":
					self.allow_lock = True
					self.sendMsg(0,"Object lock enabled")
				else:
					self.allow_lock = False
					self.sendMsg(0,"Object lock disabled")

			if cmd == "btnAdCreate" and cmdVar:
				if cmdVar == "1":
					self.allow_create = True
					self.sendMsg(0,"Object create enabled")
				else:
					self.allow_create = False
					self.sendMsg(0,"Object create disabled")

			if cmd == "delAll" and cmdVar:
				if self.p_objects.has_key(cmdVar):
					pl = p_objects[cmdVar]
					plD = pdata[cmdVar]
					plD.grabbed = []
					for o in plD.objects:
						self.deleteObject(o)
					sendMsg(pl,"All objects deleted")
			elif cmd == "delAll":
				pD.grabbed = []
				for o in pD.objects:
					self.deleteObject(o)

		if cmd == "btnNone":
			self.dbg("btnNone")

	def run_sbxTimer(self, data):
		try:
			for q in bf2.objectManager.getObjectsOfTemplate("eu_ar_rifleProjectile"):
				print "PROJECTILE DETECTED!!!AHHHHMOTHErLAND"
				if q.isValid():
					if not self.calculatedProjectiles.count(q) and q.isValid():
						foo = ["",99999]
						for p in bf2.playerManager.getPlayers():
							dist = self.getVectorDistance(p.getDefaultVehicle().getPosition(),q.getPosition())
							if dist < foo[1]:
								foo = [p,dist]
						pD = self.pdata[foo[0].getName()]
						if pD.grabbed:
							self.onRemoteCommand(foo[0].index, "btnDrop")
						else:
							self.onRemoteCommand(foo[0].index, "btnGrab")
						self.calculatedProjectiles.append(q)
				else:
					self.calculatedProjectiles.append(q)
			for q in bf2.objectManager.getObjectsOfTemplate("eu_handgun_Projectile"):
				print q.getPosition()
				if q.isValid():
					if not self.calculatedProjectiles.count(q) and q.isValid():
						foo = ["",99999]
						for p in bf2.playerManager.getPlayers():
							dist = self.getVectorDistance(p.getDefaultVehicle().getPosition(),q.getPosition())
							if dist < foo[1]:
								foo = [p,dist]
						pD = self.pdata[foo[0].getName()]
						pD.projectileGrab = q.getPosition()
						self.onRemoteCommand(foo[0].index, "btnGrab")
						self.calculatedProjectiles.append(q)
				else:
					self.calculatedProjectiles.append(q)
			self.timerCounter += 1
			if self.timerCounter >= 5:
				self.timerCounter = 0
			for p in bf2.playerManager.getPlayers():
				if not self.p_objects.has_key(p.getName()):
					self.p_objects[p.getName()] = p
				if not self.pdata.has_key(p.getName()):
					self.pdata[p.getName()] = pData()
					#self.dbg(str(p.getName()))
				#else:
				#	pl = self.p_objects[p.getName()]
				#	if not pl.isValid():
				#		self.p_objects[p.getName()] = p
			#for x in self.p_objects:
			if self.lastObject:
				oName = self.lastObject[0]
				user = self.lastObject[1]
				oPos = self.lastObject[2]
				oRot = self.lastObject[3]
				p = bf2.PlayerManager.Player(user)
				pD = self.pdata[p.getName()]
				print "-------"
				if (type(oPos) == type(('tuple','type'))) and (type(oRot) == type(('tuple','type'))) and p.isAlive():
					o = None
					for q in bf2.objectManager.getObjectsOfTemplate(str(oName)):
						if q.isValid() and not self.preCompletedObjects.count(q):
							qDist = self.getVectorDistance(p.getDefaultVehicle().getPosition(),q.getPosition())
							print qDist
							#if qDist <= (pD.tracerDistance + 10.0):
							o = q
					if o:
						self.preCompletedObjects.append(o)
						#objs = host.rcon_invoke("object.listObjectsofTemplate " + str(oName))
						#print objs
						#objs = objs.split()
						#if self.is_spGame:
						#	o.oId = objs[-7]
						o.setPosition(oPos)
						o.setRotation(oRot)
						pD.objects.append(o)
						if pD.autoGrab and (not self.allowed_vehicles.count(oName)):
							pD.grabbed = [o,pD.tracerDistance,1,o.getPosition(),o.getRotation()]
						self.lastObject = []

			for x in bf2.playerManager.getPlayers():
				p = bf2.PlayerManager.Player(x.index)
				pD = self.pdata[p.getName()]
				if pD.spawnDelay > 0:
					pD.spawnDelay -= 1
				if p.isAlive():
					#p = self.p_objects[p.index]
					pPos = p.getDefaultVehicle().getPosition()
					pRot = p.getDefaultVehicle().getRotation()

					#### Get all grabbed objects and move them where they need to be
					if pD.grabbed:
						if pD.grabbed[0].isValid():
							pCam = self.getSoldierCam(p)
							pCamRot = pCam.getRotation()
							sPos = self.getVectoredPosition(pPos,pD.grabbed[1],pRot,pCamRot)
							if pD.gridSnap < 3:
								sPos = (round(sPos[0],pD.gridSnap),round(sPos[1],pD.gridSnap),round(sPos[2],pD.gridSnap))
							if pD.grabbed[0].getPosition() != sPos:
								pD.grabbed[0].setPosition(sPos)
								self.dbg("new position: " + str(sPos[0]) + "," + str(sPos[1]) + "," + str(sPos[2]))
							pD.grabbed[0].setRotation(pD.grabbed[4])
						else:
							pD.grabbed = []

					#if self.is_spGame and pD.preGrabMode == 1 and self.timerCounter == 0:
					if pD.preGrabMode == 1 and self.timerCounter == 0:
						pCam = self.getSoldierCam(p)
						shortest = []
						shortest = [0,pD.tracerDistance + 10.0]
						pCamRot = pCam.getRotation()
						sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
						sRot = pRot
						for v in pD.objectTypes:
							if self.allowed_objects.count(v):
								for q in bf2.objectManager.getObjectsOfTemplate(v):
									if q.isValid():
										qDist = self.getVectorDistance(sPos,q.getPosition())
										if qDist < shortest[1]:
											shortest = [q,qDist]
						if pD.preGrab:
							if shortest[1] > pD.preGrab[1]:
								shortest[0] = pD.preGrab[0]
								shortest[1] = pD.preGrab[1]
						else:
							pD.preGrab = [shortest[0],shortest[1],1,shortest[0].getPosition(),pCam.getRotation()]
						if pD.preGrab[0] != shortest[0]:
							# Object, current flash thing, object position, player rotation
							o = pD.preGrab[0]
							if o.isValid():
								if hasattr(o,"oId"):
									host.rcon_invoke("Object.active id" + str(o.oId))
									host.rcon_invoke("Object.isVisible 1")
								else:
									o.setPosition(pD.preGrab[3])
								pD.preGrab[1] = 1
							pD.preGrab = [shortest[0],shortest[1],1,shortest[0].getPosition(),pCam.getRotation()]
						else:
							if pD.preGrab[1] == 1:
								o = pD.preGrab[0]
								if o.isValid():
									if hasattr(o,"oId"):
										host.rcon_invoke("Object.active id" + str(o.oId))
										host.rcon_invoke("Object.isVisible 0")
									else:
										o.setPosition((0.0,83453.0,0.0))
								pD.preGrab[1] = 2
							else:
								o = pD.preGrab[0]
								if o.isValid():
									if hasattr(o,"oId"):
										host.rcon_invoke("Object.active id" + str(o.oId))
										host.rcon_invoke("Object.isVisible 1")
									else:
										o.setPosition(pD.preGrab[3])
								pD.preGrab[1] = 1

			#### Get number of players to denote spGame
			if bf2.playerManager.getNumberOfPlayers() < 2:
				#self.is_spGame = 1
				self.is_spGame = 1
			else:
				self.is_spGame = 0
		except:
			ExceptionOutput()

		try:
			if self.timerLoaderActive:
				try:
					self.timerLoaderObjects[self.timerLoader[1]]
				except:
					try:
						#self.obj_freeze.update(self.timerLoaderTempFreeze)
						self.setAllowGrab(True)
						self.setAllowCreate(True)
						self.timerLoaderActive = 0
						self.timerLoader = [1,0,""]
						self.timerLoaderObjects = []
						self.timerLoaderTempFreeze = {}
						#self.obj_freeze = None
						#self.obj_freeze = {}
						self.sendMsg(0,"Loading finished.")
						self.dbg("Load finalized.")
						return 0
					except:
						ExceptionOutput()
				if self.timerLoader[0] == 1:
					oName = self.timerLoaderObjects[self.timerLoader[1]][0]
					oPos = self.timerLoaderObjects[self.timerLoader[1]][1]
					oRot = self.timerLoaderObjects[self.timerLoader[1]][2]
					self.timerLoader[2] = oName
					self.dbg("Created spawn template - " + str(oName))
					self.createSpawnerObject(str(oName), oPos, oRot)
					self.timerLoader[0] = 2
					#self.sendMsg(0,"Step 1 of object " + str(self.timerLoader[1]))
				elif self.timerLoader[0] == 2:
					#if not extended_objects.count(timerLoaderObjects[timerLoader[1]][0]):
					proper_obj = 0
					for q in bf2.objectManager.getObjectsOfTemplate(self.timerLoaderObjects[self.timerLoader[1]][0]):
						if q.isValid():
							proper_obj = q
					oName = self.timerLoaderObjects[self.timerLoader[1]][0]
					oPos = self.timerLoaderObjects[self.timerLoader[1]][1]
					oRot = self.timerLoaderObjects[self.timerLoader[1]][2]
					if proper_obj and not self.ragdoll_objects.count(oName.lower()):
						#self.timerLoaderTempFreeze[proper_obj] = [(float(oPos[0]),float(oPos[1]),float(oPos[2])),(float(oRot[0]),float(oRot[1]),float(oRot[2]))]
						self.obj_freeze[proper_obj] = [(float(oPos[0]),float(oPos[1]),float(oPos[2])),(float(oRot[0]),float(oRot[1]),float(oRot[2]))]
					self.timerLoader[0] = 1
					self.timerLoader[1] += 1
					#self.sendMsg(0,"Step 2 of object " + str(self.timerLoader[1]))
		except:
			ExceptionOutput()

	def freezer(self,data):
		#### Refreeze all frozen objects
		invalidFreezes = []
		for o in self.obj_freeze:
			if self.obj_freeze[o]:
				if (o.isValid()) and (type(self.obj_freeze[o][0]) == type(('tuple','thing'))):
					if o.getPosition() != self.obj_freeze[o][0]:
						o.setPosition(self.obj_freeze[o][0])
				else:
					invalidFreezes.append(o)
		for f in invalidFreezes:
			if self.obj_freeze.has_key(f):
				self.obj_freeze.pop(f)

	def createSpawnerObject(self, sName, sPos, sRot = (0.0,0.0,0.0), user=0):
		if not self.spawnerTemplates.count(sName):
			host.rcon_invoke("ObjectTemplate.create ObjectSpawner sbx_" + sName)
			host.rcon_invoke("ObjectTemplate.activeSafe ObjectSpawner sbx_" + sName)
			host.rcon_invoke("ObjectTemplate.hasMobilePhysics 0")
			host.rcon_invoke("ObjectTemplate.setObjectTemplate 1 " + sName)
			host.rcon_invoke("ObjectTemplate.setObjectTemplate 2 " + sName)
			host.rcon_invoke("ObjectTemplate.minSpawnDelay 700000")
			host.rcon_invoke("ObjectTemplate.maxSpawnDelay 900000")
			self.spawnerTemplates.append(sName)
		host.rcon_invoke("Object.create sbx_" + sName)
		host.rcon_invoke("Object.absolutePosition " + str(sPos[0]) + "/" + str(sPos[1]) + "/" + str(sPos[2]))
		host.rcon_invoke("Object.rotation " + str(sRot[0]) + "/" + str(sRot[1]) + "/" + str(sRot[2]))
		host.rcon_invoke("Object.team 2")
		#host.rcon_invoke("Object.delete")
		self.lastObject = [str(sName),user,sPos,sRot]

	def createSpawnerTempl(self, id, tmpl):
		host.rcon_invoke("ObjectTemplate.create ObjectSpawner sbx_" + id)
		host.rcon_invoke("ObjectTemplate.activeSafe ObjectSpawner sbx_" + id)
		host.rcon_invoke("ObjectTemplate.hasMobilePhysics 0")
		host.rcon_invoke("ObjectTemplate.setObjectTemplate 1 " + tmpl)
		host.rcon_invoke("ObjectTemplate.setObjectTemplate 2 " + tmpl)
		host.rcon_invoke("ObjectTemplate.minSpawnDelay 700000")
		host.rcon_invoke("ObjectTemplate.maxSpawnDelay 900000")

	def createSpawner(self, id, sPos, sRot):
		host.rcon_invoke("Object.create sbx_" + id)
		host.rcon_invoke("Object.absolutePosition " + str(sPos[0]) + "/" + str(sPos[1]) + "/" + str(sPos[2]))
		host.rcon_invoke("Object.rotation " + str(sRot[0]) + "/" + str(sRot[1]) + "/" + str(sRot[2]))
		host.rcon_invoke("Object.team 2")
		host.rcon_invoke("Object.delete")

	def onChatMessage(self, playerId, text, channel, flags):
		text = text.replace( "HUD_TEXT_CHAT_TEAM", "" )
		text = text.replace( "HUD_TEXT_CHAT_SQUAD", "" )
		text = text.replace( "HUD_CHAT_DEADPREFIX", "" )
		text = text.replace( "*\xA71DEAD\xA70*", "" )
		if text[0:1] == "!":
			pattern = re.compile(r'!(\w*) ?(.*)')
			matches = pattern.findall(text)
			command = matches[0][0]
			if matches[0][1] != "":
				parameter = matches[0][1]
			else:
				parameter = None
			if playerId == -1: #fix for local servers
				playerId = 255
			p = bf2.PlayerManager.Player(playerId)
			pD = self.pdata[p.getName()]
			if command == "r" and parameter:
				self.onRemoteCommand(playerId, parameter)
			if self.MP_LS_USERS.count(p.getName()) or self.is_spGame == 1:
				if command == "save" and parameter:
					self.saveSandbox(parameter)
				if command == "load" and parameter:
					if self.is_spGame:
						self.loadSandbox(parameter,p)
					else:
						self.advloadSandbox(parameter,playerId)
				if command == "clear":
					host.rcon_invoke("admin.restartmap")

	def saveSandbox(self, filename):
		self.sendMsg(0, "Oops, saving doesn't work yet in 2142. Sorry.")
		return 0;
		file_string = str(host.sgl_getMapName()) + "," + str(self.SANDBOX_VERSION) + "@"
		out_file = open("mods/sbx2142/saved/" + str(filename) + "_" + str(host.sgl_getMapName()).lower() + ".sbx","w")
		increment = 0
		for t in self.object_types:
			for q in bf2.objectManager.getObjectsOfTemplate(t):
				if q.isValid():
					if not self.deleted_objects.count(q):
						increment += 1
						qPos = q.getPosition()
						qRot = q.getRotation()
						qNam = q.templateName
						file_string += str(qNam) + ";" + str(qPos[0]) + "," + str(qPos[1]) + "," + str(qPos[2]) + ";" + str(qRot[0]) + "," + str(qRot[1]) + "," + str(qRot[2]) + "!"
		out_file.write(file_string)
		out_file.close()
		self.sendMsg(0, "Your sandbox has been saved as mods/sbx2142/saved/" + str(filename) + "_" + str(host.sgl_getMapName()).lower() + ".sbx")
		self.dbg("Saved file, " + str(increment) + " objects")

	def loadSandbox(self, filename, p):
		pD = self.pdata[p.getName()]
		try: in_file = open("mods/sbx2142/saved/" + str(filename) + "_" + str(host.sgl_getMapName()).lower() + ".sbx","r")
		except:
			self.sendMsg(0, "The file you specified was not found.")
			return 0
		savedStr = in_file.read()
		a = savedStr.split("@")
		w = a[0].split(",")
		if not w.count(self.SANDBOX_VERSION):
			self.sendMsg(0, "This file was created with a different version of BF2:Sandbox. Some objects may be loaded incorrectly.")
		if w[0].lower() == str(host.sgl_getMapName()).lower():
			b = a[1].split("!")
			increment = 0
			for objInfo in b:
				if objInfo:
					objInfoList = objInfo.split(";")
					oName = objInfoList[0]
					oPos = objInfoList[1].split(",")
					oRot = objInfoList[2].split(",")
					if self.mp_objects.count(oName.lower()) and not self.is_spGame:
						oName += "_mp"
					#self.createObject(oName, oPos, oRot)
					self.createSpawnerObject(oName, oPos, oRot)
					if pD.objectTypes.count(oName) < 1:
						pD.objectTypes.append(oName)
					if self.object_types.count(oName) < 1:
						self.object_types.append(oName)
					increment += 1
			self.sendMsg(0, str(filename) + ".bf2sbx has been loaded.")
			self.dbg("Loaded file, " + str(increment) + " objects")
		else:
			self.sendMsg(0, "The file you specified was not created in this map.")
		in_file.close()

	def advloadSandbox(self, filename, playerId):
		p = bf2.PlayerManager.Player(playerId)
		pD = pdata[p.getName()]
		try: in_file = open("mods/sandbox/saved/" + str(filename) + ".bf2sbx","r")
		except:
			self.sendMsg(0,"The file you specified was not found.")
			return 0
		savedStr = in_file.read()
		a = savedStr.split("@")
		w = a[0].split(",")
		if not w.count(self.SANDBOX_VERSION):
			self.sendMsg(0, "This file was created with a different version of BF2:Sandbox. Some objects may be loaded incorrectly.")
		if not self.is_spGame:
			self.sendMsg(0, "Loading a sandbox into a multiplayer game may cause unpredictable behavior.")
		if w[0].lower() == str(host.sgl_getMapName()).lower():
			self.timerLoaderActive = 0
			self.timerLoader = [1,0,""]
			self.timerLoaderObjects = []
			b = a[1].split("!")
			self.dbg("Starting object loops")
			increment = 0
			for objInfo in b:
				if objInfo:
					self.dbg("Found an object")
					objInfoList = objInfo.split(";")
					oName = objInfoList[0]
					oPos = objInfoList[1].split(",")
					oRot = objInfoList[2].split(",")
					oPos = (float(oPos[0]),float(oPos[1]),float(oPos[2]))
					oRot = (float(oRot[0]),float(oRot[1]),float(oRot[2]))
					if self.mp_objects.count(oName.lower()) and not self.is_spGame:
						oName += "_mp"
					self.timerLoaderObjects.append([oName,oPos,oRot])
					if pD.objectTypes.count(oName) < 1:
						pD.objectTypes.append(oName)
					if self.object_types.count(oName) < 1:
						self.object_types.append(oName)
					increment += 1
			self.sendMsg(0,str(filename) + ".bf2sbx loading using experimental MP mode")
			self.setAllowGrab(False)
			self.setAllowCreate(False)
			self.timerLoaderActive = 1
		else:
			self.sendMsg(0,"The file you specified was not created in this map.")
		in_file.close()

	def fixPlayerObject(self, pId, pName):
		try:
			old = self.p_objects[pName]
			new = bf2.PlayerManager.Player(pId)

			new.tracerDistance = old.tracerDistance
			new.objects = old.objects
			new.objectTypes = old.objectTypes
			new.grabbed = old.grabbed
			new.lastRotation = old.lastRotation
			new.lastSpawned = old.lastSpawned
			new.lastTemplate = old.lastTemplate
			new.tracerOn = old.tracerOn
			new.tracerDistance = old.tracerDistance
			new.gridSnap = old.gridSnap
			new.rotSnap = old.rotSnap
			new.autoGrab = old.autoGrab
			new.numSpawned = old.numSpawned
			new.preGrab = old.preGrab
			new.preGrabMode = old.preGrabMode
			new.lastAction = old.lastAction
			new.spawnDelay = old.spawnDelay

			self.p_objects[pName] = new
		except:
			ExceptionOutput()

	def onEnterVehicle(self, player, vehicle, freeSoldier = False):
		pD = self.pdata[player.getName()]
		pD.grabbed = []

	def onExitVehicle(self, player, vehicle):
		pD = self.pdata[player.getName()]
		pD.grabbed = []

	def onPlayerDeathCQ(self, victim, vehicle):
		pD = self.pdata[victim.getName()]
		pD.grabbed = []

	def onPlayerKilledCQ(self, victim, attacker, weapon, assists, object):
		pD = self.pdata[victim.getName()]
		pD.grabbed = []

	def onPlayerDisconnect(self, player):
		if self.pdata.has_key(player.getName()):
			pD = pdata[player.getName()]
			pD.grabbed = []
			for o in pD.objects:
				self.deleteObject(o)
		for o in self.obj_locked:
			if self.obj_locked[o] == player.index:
				self.obj_locked.pop(o)
		try:
			self.pdata.pop(player.getName())
		except:
			ExceptionOutput()

	def getVectoredPosition(self, pPos, distance, pRot, pCamRot):
		from math import pi
		#Calculate the height using a triangle
		p = pCamRot[1] #Pitch of the player's cam
		a = 90.0 - abs(p)
		p = pi * p / 180.0
		a = pi * a / 180.0
		sVz = (distance * math.sin(p))/math.sin(a)
		#convert from crazy-ass BF2 values to polar coordinate values
		new_pRot = -(pRot[0] + pCamRot[0]) + 90.0
		sVx,sVy = self.rect(distance,new_pRot,1) #Convert vector of [spawnDistance] with heading of cam rotation to x,y values
		sPos = (pPos[0] + sVx, pPos[1] - sVz, pPos[2] + sVy)
		return sPos

	def onPlayerSpawn(self, player, soldier):
		self.dbg(player.isFlagHolder())
		self.dbg(player.getAddress())
		self.dbg(player.isRemote())
		#if not self.p_objects.has_key(player.index):
		#self.p_objects[player.index] = player
		#self.p_objects[player.getName()] = bf2.PlayerManager.Player(player.index)
		#self.dbg("Set a pobject for " + str(player.getName()))
		#pass #Do nothing

	def rect(self, r, w, deg=0): # radian if deg=0; degree if deg=1
		from math import cos, sin, pi
		if deg:
			w = pi * w / 180.0
		return r * cos(w), r * sin(w)

	def sendMsg(self, p, msg):
		if p:
			host.rcon_invoke("game.sayAll \"" + str(p.getName()) + " - " + str(msg) + "\"")
		else:
			host.rcon_invoke("game.sayAll \"" + str(msg) + "\"")

	def getVectorDistance(self, pos1, pos2):
		diffVec = [0.0, 0.0, 0.0]
		diffVec[0] = math.fabs(pos1[0] - pos2[0])
		diffVec[1] = math.fabs(pos1[1] - pos2[1])
		diffVec[2] = math.fabs(pos1[2] - pos2[2])
		return math.sqrt(diffVec[0] * diffVec[0] + diffVec[1] * diffVec[1] + diffVec[2] * diffVec[2])

	def getSoldierCam(self, player):
		p = player.getDefaultVehicle()
		kids = p.getChildren()
		for child in kids:
			if child.templateName == "SoldierCamera":
				return child

	def disableBoundaries(self):
		host.rcon_invoke("CombatAreaManager.use 0")
		host.rcon_invoke("CombatAreaManager.timeAllowedOutside 1000.000000")
		self.sBoundaries = False

	def enableBoundaries(self):
		host.rcon_invoke("CombatAreaManager.use 1")
		host.rcon_invoke("CombatAreaManager.timeAllowedOutside 10.00")
		self.sBoundaries = True

	def setAllowedObjects(self,objList):
		if type(self.allowed_objects) == type(objList):
			self.allowed_objects = objList

	def addAllowedObjects(self,obj):
		# This can take either a list or a string.
		if type(self.allowed_objects) == type(obj):
			self.allowed_objects.extend(obj)
		else:
			self.allowed_objects.append(obj)

	def delAllowedObjects(self,obj):
		# This can take either a list or a string.
		if type(self.allowed_objects) == type(obj):
			for x in obj:
				try:
					self.allowed_objects.remove(x)
				except:
					pass #do nothing
		else:
			self.allowed_objects.remove(obj)

	def setSpawnLimits(self,objSp,ragdSp):
		self.SPAWN_LIMIT = objSp
		self.RAGDOLL_LIMIT = ragdSp

	def setMpAdmins(self,adminList):
		if type(self.allowed_objects) == type(adminList):
			self.MP_LS_USERS = adminList
		else:
			self.MP_LS_USERS = [adminList]

	def setAllowLock(self,val):
		if val:
			self.allow_lock = True
		else:
			self.allow_lock = False

	def setAllowGrab(self,val):
		if val:
			self.allow_grab = True
		else:
			self.allow_grab = False

	def setAllowCreate(self,val,override=True):
		if val:
			self.allow_create = True
		else:
			self.allow_create = False
		if override:
			self.allowCreateOverride = True
		else:
			self.allowCreateOverride = False

	def setMinPlayers(self,val):
		host.rcon_invoke("sv.numPlayersNeededToStart " + str(val))

	def createObject(self, oName, oPos, oRot = (0.0,0.0,0.0), user = 0):
		try:
			oPos = (float(oPos[0]),float(oPos[1]),float(oPos[2]))
			oRot = (float(oRot[0]),float(oRot[1]),float(oRot[2]))
			host.rcon_invoke("gameLogic.createObject " + str(oName))
			print "host.rcon_invoke(\"gameLogic.createObject " + str(oName) + "\""
			self.lastObject = [str(oName),user,oPos,oRot]
			#for q in bf2.objectManager.getObjectsOfTemplate(str(oName)):
			#	if q.isValid():
			#		lastObj = q
			#objs = host.rcon_invoke("object.listObjectsofTemplate " + str(oName))
			#objs = objs.split()
			#lastObj.oId = objs[-7]
			#lastObj.setPosition(oPos)
			#lastObj.setRotation(oRot)
			#return lastObj
			return 1
		except:
			ExceptionOutput()

	def deleteObject(self, o):
		try:
			if o:
				if o.isValid():
					if hasattr(o,"oId"):
						host.rcon_invoke("Object.active id" + str(o.oId))
						host.rcon_invoke("Object.delete")
					else:
						o.setPosition((0.0,99999.0,0.0))
						self.deleted_objects.append(o)
		except:
			ExceptionOutput()

def start():
	return sandboxCore()

print "Sandbox core loaded."
